#  Copyright 2019 United Kingdom Research and Innovation
#  Copyright 2019 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt


from cil.framework import Processor, DataContainer, AcquisitionData,\
 AcquisitionGeometry, ImageGeometry, ImageData
import numpy
import warnings

class Normaliser(Processor):
    r""" This processor can be used to normalise data with flat-field and dark-field images.

    The dark-field image is used to remove the offset from the input data, and the flat-field image is used to scale the data. 

    The normalisation is done as follows:

    .. math::
        \text{output} = \frac{\text{input} - \text{offset}}{\text{scale} - \text{offset}}

    where :math:`\text{input}`, :math:`\text{offset}` and :math:`\text{scale}` are the input, dark-field and flat-field images respectively.

    The processor can be used in-place to reduce memory usage.

    Parameters
    ----------
    flat_field : numpy.ndarray, optional
        The flat field image. It must have the same shape as a single projection.
    dark_field : numpy.ndarray, optional
        The dark field image. It must have the same shape as a single projection.
    tolerance : float, optional
        The value to substitute for output values where division by zero occurs.

        
    Examples
    --------
    Basic usage with flat and dark fields:

    >>> from cil.processors import Normaliser
    >>> normaliser = Normaliser(flat_field, dark_field)
    >>> normaliser.set_input(data)
    >>> normalised_data = normaliser.get_output()

    In-place usage:

    >>> from cil.processors import Normaliser
    >>> normaliser = Normaliser(flat_field, dark_field)
    >>> normaliser.set_input(data)
    >>> normaliser.get_output(out=data)

    """

    def __init__(self, flat_field = None, dark_field = None, tolerance = 1e-5):
        kwargs = {
                  'flat_field'  : flat_field,
                  'dark_field'  : dark_field,
                  # very small number. Used when there is a division by zero
                  'tolerance'   : tolerance
                  }

        #DataProcessor.__init__(self, **kwargs)
        super(Normaliser, self).__init__(**kwargs)
        if not flat_field is None:
            self.set_flat_field(flat_field)
        if not dark_field is None:
            self.set_dark_field(dark_field)

    def check_input(self, dataset):
        if dataset.number_of_dimensions == 3 or\
           dataset.number_of_dimensions == 2:
               return True
        else:
            raise ValueError("Expected input dimensions is 2 or 3, got {0}"\
                             .format(dataset.number_of_dimensions))

    def set_dark_field(self, df):
        if type(df) is numpy.ndarray:
            if len(numpy.shape(df)) == 3:
                raise ValueError('Dark Field should be 2D')
            elif len(numpy.shape(df)) == 2:
                self.dark_field = df
        elif issubclass(type(df), DataContainer):
            self.dark_field = df.as_array()

    def set_flat_field(self, df):
        if type(df) is numpy.ndarray:
            if len(numpy.shape(df)) == 3:
                raise ValueError('Flat Field should be 2D')
            elif len(numpy.shape(df)) == 2:
                self.flat_field = df
        elif issubclass(type(df), DataContainer):
            self.flat_field = df.as_array()


    @staticmethod
    def estimate_normalised_error(projection, flat, dark, delta_flat, delta_dark):
        '''returns the estimated relative error of the normalised projection

        n = (projection - dark) / (flat - dark)
        Dn/n = (flat-dark + projection-dark)/((flat-dark)*(projection-dark))*(Df/f + Dd/d)
        '''
        a = (projection - dark)
        b = (flat-dark)
        df = delta_flat / flat
        dd = delta_dark / dark
        rel_norm_error = (b + a) / (b * a) * (df + dd)
        return rel_norm_error

    def process(self, out=None):

        input_arr = self.get_input().array

        if out is None:
            out = self.get_input().geometry.allocate(None)
        
        out_array = out.array

        if self.flat_field is None and self.dark_field is None:
            raise ValueError('No flat or dark field provided. Please provide at least one.')
        
        if self.dark_field is None:
            offset = 0
        else:
            if input_arr.shape[-len(self.dark_field.shape)::]!=self.dark_field.shape:
                raise ValueError('Dark field and projections size do not match.')
            offset = self.dark_field

        zero_err = False
        if self.flat_field is not None:
            if input_arr.shape[-len(self.flat_field.shape)::]!=self.flat_field.shape:
                raise ValueError('Flat field and projections size do not match.')
            
            scale = self.flat_field - offset

            numpy.seterr(divide='warn')
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")

                denom_inv = 1.0 / scale
                # Check if any warnings were raised and handle these later
                if len(w) > 0 and issubclass(w[-1].category, RuntimeWarning):
                    zero_err = True
        
        if self.dark_field is not None:
            numpy.subtract(input_arr, offset, out=out_array)
            input_arr = out_array

        if self.flat_field is not None:
            numpy.multiply(input_arr, denom_inv, out=out_array)

        if zero_err:
            warnings.warn("Divide by zero detected. These values will be substituted by the set tolerance. Please check your data for bad pixels.", UserWarning, stacklevel=2)
            # create map of zeros in scale array
            zero_map = numpy.where(scale == 0)
            # set these to the tolerance value in all projections
            for i in range(out_array.shape[0]):
                out_array[i][zero_map] = self.tolerance

        return out



#  Copyright 2022 United Kingdom Research and Innovation
#  Copyright 2022 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

import unittest
from cil.framework import AcquisitionGeometry
import numpy as np
import math

from utils import has_astra, initialise_tests
initialise_tests()

if has_astra:
    from cil.plugins.astra.utilities import convert_geometry_to_astra
    from cil.plugins.astra.utilities import convert_geometry_to_astra_vec_2D
    from cil.plugins.astra.utilities import convert_geometry_to_astra_vec_3D

from utils_projectors import create_cone_flex_default_ig


class TestGeometry_Parallel2D(unittest.TestCase):
    def setUp(self):
        self.pixels_x = 128

        angles_deg = np.asarray([0,90.0,180.0], dtype='float32')
        angles_rad = angles_deg * np.pi /180.0

        self.ag = AcquisitionGeometry.create_Parallel2D()\
                                .set_angles(angles_rad, angle_unit='radian')\
                                .set_labels(['angle','horizontal'])\
                                .set_panel(self.pixels_x, 0.1)

        self.ig = self.ag.get_ImageGeometry()

        self.ag_deg = AcquisitionGeometry.create_Parallel2D()\
                                    .set_angles(angles_deg, angle_unit='degree')\
                                    .set_labels(['angle','horizontal'])\
                                    .set_panel(self.pixels_x, 0.1)


    @unittest.skipUnless(has_astra, "Requires ASTRA")
    def test_convert_geometry_simple(self):

        #2D parallel radians
        astra_vol, astra_sino = convert_geometry_to_astra(self.ig, self.ag)

        self.assertEqual(astra_sino['type'],  'parallel')
        self.assertEqual(astra_sino['DetectorCount'], self.ag.pixel_num_h)
        self.assertEqual(astra_sino['DetectorWidth'], self.ag.pixel_size_h)
        np.testing.assert_allclose(astra_sino['ProjectionAngles'], -self.ag.angles)

        #2D parallel degrees
        astra_vol, astra_sino = convert_geometry_to_astra(self.ig, self.ag_deg)
        np.testing.assert_allclose(astra_sino['ProjectionAngles'], -self.ag.angles)

        # check the image geometry
        self.assertEqual(astra_vol['GridColCount'], self.ig.voxel_num_x)
        self.assertEqual(astra_vol['GridRowCount'], self.ig.voxel_num_y)
        self.assertEqual(astra_vol['option']['WindowMinX'], -self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxX'], self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMinY'], -self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxY'], self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)


    @unittest.skipUnless(has_astra, "Requires ASTRA")
    def test_convert_geometry_vector2D(self):

        # 2D parallel radians
        astra_vol, astra_sino = convert_geometry_to_astra_vec_2D(self.ig, self.ag)

        self.assertEqual(astra_sino['type'],  'parallel_vec')
        self.assertEqual(astra_sino['DetectorCount'], self.ag.pixel_num_h)

        # note the sign of the y values differs from the astra 3D projector case
        vectors = np.zeros((3,6),dtype='float64')

        vectors[0][1] = -1.0
        vectors[0][4] = self.ag.pixel_size_h

        vectors[1][0] = 1.0
        vectors[1][5] = self.ag.pixel_size_h

        vectors[2][1] = 1.0
        vectors[2][4] = -self.ag.pixel_size_h

        np.testing.assert_allclose(astra_sino['Vectors'], vectors, atol=1e-6)

        # 2D parallel degrees
        astra_vol, astra_sino = convert_geometry_to_astra_vec_2D(self.ig, self.ag_deg)
        np.testing.assert_allclose(astra_sino['Vectors'], vectors, atol=1e-6)

        # image geometry
        self.assertEqual(astra_vol['GridColCount'], self.ig.voxel_num_x)
        self.assertEqual(astra_vol['GridRowCount'], self.ig.voxel_num_y)

        self.assertEqual(astra_vol['option']['WindowMinX'], -self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxX'], self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMinY'], -self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxY'], self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)


    @unittest.skipUnless(has_astra, "Requires ASTRA")
    def test_convert_geometry_vector3D(self):

        # 2D parallel radians
        astra_vol, astra_sino = convert_geometry_to_astra_vec_3D(self.ig, self.ag)

        self.assertEqual(astra_sino['type'],  'parallel3d_vec')
        self.assertEqual(astra_sino['DetectorRowCount'], 1.0)
        self.assertEqual(astra_sino['DetectorColCount'], self.ag.pixel_num_h)

        vectors = np.zeros((3,12),dtype='float64')

        vectors[0][1] = 1.0
        vectors[0][6] = self.ag.pixel_size_h
        vectors[0][11] = self.ag.pixel_size_h

        vectors[1][0] = 1.0
        vectors[1][7] = -self.ag.pixel_size_h
        vectors[1][11] = self.ag.pixel_size_h

        vectors[2][1] = -1.0
        vectors[2][6] = -self.ag.pixel_size_h
        vectors[2][11] = self.ag.pixel_size_h

        np.testing.assert_allclose(astra_sino['Vectors'], vectors, atol=1e-6)

        # 2D parallel degrees
        astra_vol, astra_sino = convert_geometry_to_astra_vec_3D(self.ig, self.ag_deg)
        np.testing.assert_allclose(astra_sino['Vectors'], vectors, atol=1e-6)

        # image geometry
        self.assertEqual(astra_vol['GridColCount'], self.ig.voxel_num_x)
        self.assertEqual(astra_vol['GridRowCount'], self.ig.voxel_num_y)
        self.assertEqual(astra_vol['GridSliceCount'], 1)

        self.assertEqual(astra_vol['option']['WindowMinX'], -self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxX'], self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMinY'], -self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxY'], self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)
        self.assertEqual(astra_vol['option']['WindowMinZ'], - self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxZ'], + self.ig.voxel_size_x * 0.5)

class TestGeometry_Parallel3D(unittest.TestCase):
    def setUp(self):
        # Define image geometry.
        pixels_x = 128
        pixels_y = 3

        angles_deg = np.asarray([0,90.0,180.0], dtype='float32')
        angles_rad = angles_deg * np.pi /180.0

        self.ag = AcquisitionGeometry.create_Parallel3D()\
                                 .set_angles(angles_rad, angle_unit='radian')\
                                 .set_labels(['vertical', 'angle','horizontal'])\
                                 .set_panel((pixels_x,pixels_y), (0.1,0.2))
        
        self.ag_deg = AcquisitionGeometry.create_Parallel3D()\
                                    .set_angles(angles_deg, angle_unit='degree')\
                                    .set_labels(['vertical', 'angle','horizontal'])\
                                    .set_panel((pixels_x,pixels_y), (0.1,0.2))
        
        self.ig = self.ag.get_ImageGeometry()


    @unittest.skipUnless(has_astra, "Requires ASTRA")
    def test_convert_geometry_simple(self):

        #3D parallel
        astra_vol, astra_sino = convert_geometry_to_astra(self.ig, self.ag)
        self.assertEqual(astra_sino['type'],  'parallel3d')
        self.assertEqual(astra_sino['DetectorColCount'], self.ag.pixel_num_h)
        self.assertEqual(astra_sino['DetectorRowCount'], self.ag.pixel_num_v)
        self.assertEqual(astra_sino['DetectorSpacingX'], self.ag.pixel_size_h)
        self.assertEqual(astra_sino['DetectorSpacingY'], self.ag.pixel_size_v)
        np.testing.assert_allclose(astra_sino['ProjectionAngles'], -self.ag.angles)

        #2D parallel degrees
        astra_vol, astra_sino = convert_geometry_to_astra(self.ig, self.ag_deg)
        np.testing.assert_allclose(astra_sino['ProjectionAngles'], -self.ag.angles)

        # image geometry
        self.assertEqual(astra_vol['GridColCount'], self.ig.voxel_num_x)
        self.assertEqual(astra_vol['GridRowCount'], self.ig.voxel_num_y)
        self.assertEqual(astra_vol['GridSliceCount'], self.ig.voxel_num_z)

        self.assertEqual(astra_vol['option']['WindowMinX'], -self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxX'], self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMinY'], -self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxY'], self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)
        self.assertEqual(astra_vol['option']['WindowMinZ'], -self.ig.voxel_num_z * self.ig.voxel_size_z * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxZ'], self.ig.voxel_num_z * self.ig.voxel_size_z * 0.5)

    @unittest.skipUnless(has_astra, "Requires ASTRA")
    def test_convert_geometry_vector2D(self):

        astra_vol, astra_sino = convert_geometry_to_astra_vec_2D(self.ig, self.ag)
        self.assertEqual(astra_sino['type'],  'parallel_vec')
        self.assertEqual(astra_sino['DetectorCount'], self.ag.pixel_num_h)

        vectors = np.zeros((3,6),dtype='float64')

        vectors[0][1] = -1.0
        vectors[0][4] = self.ag.pixel_size_h

        vectors[1][0] = 1.0
        vectors[1][5] = self.ag.pixel_size_h

        vectors[2][1] = 1.0
        vectors[2][4] = -self.ag.pixel_size_h

        np.testing.assert_allclose(astra_sino['Vectors'], vectors, atol=1e-6)

        #degrees
        astra_vol, astra_sino = convert_geometry_to_astra_vec_2D(self.ig, self.ag_deg)
        np.testing.assert_allclose(astra_sino['Vectors'], vectors, atol=1e-6)

        # image geometry
        self.assertEqual(astra_vol['GridColCount'], self.ig.voxel_num_x)
        self.assertEqual(astra_vol['GridRowCount'], self.ig.voxel_num_y)

        self.assertEqual(astra_vol['option']['WindowMinX'], -self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxX'], self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMinY'], -self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxY'], self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)


    @unittest.skipUnless(has_astra, "Requires ASTRA")
    def test_convert_geometry_vector3D(self):

        astra_vol, astra_sino = convert_geometry_to_astra_vec_3D(self.ig, self.ag)
        self.assertEqual(astra_sino['type'],  'parallel3d_vec')
        self.assertEqual(astra_sino['DetectorRowCount'], self.ag.pixel_num_v)
        self.assertEqual(astra_sino['DetectorColCount'], self.ag.pixel_num_h)


        #ray : the ray direction
        #d : the center of the detector
        #u : the vector from detector pixel (0,0) to (0,1)
        #v : the vector from detector pixel (0,0) to (1,0)

        vectors = np.zeros((3,12),dtype='float64')

        vectors[0][1] = 1.0
        vectors[0][6] = self.ag.pixel_size_h
        vectors[0][11] = self.ag.pixel_size_v

        vectors[1][0] = 1.0
        vectors[1][7] = -self.ag.pixel_size_h
        vectors[1][11] = self.ag.pixel_size_v

        vectors[2][1] = -1.0
        vectors[2][6] = -self.ag.pixel_size_h
        vectors[2][11] = self.ag.pixel_size_v

        np.testing.assert_allclose(astra_sino['Vectors'], vectors, atol=1e-6)

        #degrees
        astra_vol, astra_sino = convert_geometry_to_astra_vec_3D(self.ig, self.ag_deg)
        np.testing.assert_allclose(astra_sino['Vectors'], vectors, atol=1e-6)

        # image geometry
        self.assertEqual(astra_vol['GridColCount'], self.ig.voxel_num_x)
        self.assertEqual(astra_vol['GridRowCount'], self.ig.voxel_num_y)
        self.assertEqual(astra_vol['GridSliceCount'], self.ig.voxel_num_z)

        self.assertEqual(astra_vol['option']['WindowMinX'], -self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxX'], self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMinY'], -self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxY'], self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)
        self.assertEqual(astra_vol['option']['WindowMinZ'], -self.ig.voxel_num_z * self.ig.voxel_size_z * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxZ'], self.ig.voxel_num_z * self.ig.voxel_size_z * 0.5)

class TestGeometry_Cone2D(unittest.TestCase):
    def setUp(self):
        pixels_x = 128

        angles_deg = np.asarray([0,90.0,180.0], dtype='float32')
        angles_rad = angles_deg * np.pi /180.0


        self.ag = AcquisitionGeometry.create_Cone2D([0,-2], [0,1])\
                                     .set_angles(angles_rad, angle_unit='radian')\
                                     .set_labels(['angle','horizontal'])\
                                     .set_panel(pixels_x, 0.1)

        self.ig = self.ag.get_ImageGeometry()

        self.ag_deg = AcquisitionGeometry.create_Cone2D([0,-2], [0,1])\
                                     .set_angles(angles_deg, angle_unit='degree')\
                                     .set_labels(['angle','horizontal'])\
                                     .set_panel(pixels_x, 0.1)
        


    @unittest.skipUnless(has_astra, "Requires ASTRA")
    def test_convert_geometry_simple(self):
        astra_vol, astra_sino = convert_geometry_to_astra(self.ig, self.ag)

        self.assertEqual(astra_sino['type'], 'fanflat')
        self.assertEqual(astra_sino['DistanceOriginSource'], self.ag.dist_source_center)
        self.assertTrue(astra_sino['DistanceOriginDetector'], self.ag.dist_center_detector)

        self.assertEqual(astra_sino['DetectorCount'], self.ag.pixel_num_h)
        self.assertEqual(astra_sino['DetectorWidth'], self.ag.pixel_size_h)
        np.testing.assert_allclose(astra_sino['ProjectionAngles'], -self.ag.angles)

        #2D parallel degrees
        astra_vol, astra_sino = convert_geometry_to_astra(self.ig, self.ag_deg)
        np.testing.assert_allclose(astra_sino['ProjectionAngles'], -self.ag.angles)

        # check the image geometry
        self.assertEqual(astra_vol['GridColCount'], self.ig.voxel_num_x)
        self.assertEqual(astra_vol['GridRowCount'], self.ig.voxel_num_y)
        self.assertEqual(astra_vol['option']['WindowMinX'], -self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxX'], self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMinY'], -self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxY'], self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)

    @unittest.skipUnless(has_astra, "Requires ASTRA")
    def test_convert_geometry_vector2D(self):
        #2D cone
        astra_vol, astra_sino = convert_geometry_to_astra_vec_2D(self.ig, self.ag)

        self.assertEqual(astra_sino['type'], 'fanflat_vec')

        vectors = np.zeros((3,6),dtype='float64')

        # note the sign of the y values differs from the astra 3D projector case

        pixel_size_v = self.ig.voxel_size_x * self.ag.magnification
        vectors[0][1] = 1 * self.ag.dist_source_center
        vectors[0][3] = -self.ag.dist_center_detector
        vectors[0][4] = self.ag.pixel_size_h

        vectors[1][0] = -1 * self.ag.dist_source_center
        vectors[1][2] = self.ag.dist_center_detector
        vectors[1][5] = self.ag.pixel_size_h

        vectors[2][1] = -self.ag.dist_source_center
        vectors[2][3] = 1 * self.ag.dist_center_detector
        vectors[2][4] = -1 * self.ag.pixel_size_h

        np.testing.assert_allclose(astra_sino['Vectors'], vectors, atol=1e-6)

        #2D parallel degrees
        astra_vol, astra_sino = convert_geometry_to_astra_vec_2D(self.ig, self.ag_deg)
        np.testing.assert_allclose(astra_sino['Vectors'], vectors, atol=1e-6)

        # check image geometry
        self.assertEqual(astra_vol['GridColCount'], self.ig.voxel_num_x)
        self.assertEqual(astra_vol['GridRowCount'], self.ig.voxel_num_y)
        self.assertEqual(astra_vol['option']['WindowMinX'], -self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxX'], self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMinY'], -self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxY'], self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)

    @unittest.skipUnless(has_astra, "Requires ASTRA")
    def test_convert_geometry_vector3D(self):
        #2D cone
        astra_vol, astra_sino = convert_geometry_to_astra_vec_3D(self.ig, self.ag)

        self.assertEqual(astra_sino['type'], 'cone_vec')

        vectors = np.zeros((3,12),dtype='float64')

        pixel_size_v = self.ig.voxel_size_x * self.ag.magnification
        vectors[0][1] = -1 * self.ag.dist_source_center
        vectors[0][4] = self.ag.dist_center_detector
        vectors[0][6] = self.ag.pixel_size_h
        vectors[0][11] = pixel_size_v

        vectors[1][0] = -1 * self.ag.dist_source_center
        vectors[1][3] = self.ag.dist_center_detector
        vectors[1][7] = -self.ag.pixel_size_h
        vectors[1][11] = pixel_size_v

        vectors[2][1] = self.ag.dist_source_center
        vectors[2][4] = -1 * self.ag.dist_center_detector
        vectors[2][6] = -1 * self.ag.pixel_size_h
        vectors[2][11] = pixel_size_v

        np.testing.assert_allclose(astra_sino['Vectors'], vectors, atol=1e-6)

        #2D parallel degrees
        astra_vol, astra_sino = convert_geometry_to_astra_vec_3D(self.ig, self.ag_deg)
        np.testing.assert_allclose(astra_sino['Vectors'], vectors, atol=1e-6)

        # check image geometry
        self.assertEqual(astra_vol['GridColCount'], self.ig.voxel_num_x)
        self.assertEqual(astra_vol['GridRowCount'], self.ig.voxel_num_y)
        self.assertEqual(astra_vol['option']['WindowMinX'], -self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxX'], self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMinY'], -self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxY'], self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)


class TestGeometry_Cone3D(unittest.TestCase):
    def setUp(self):
        pixels_x = 128
        pixels_y = 3

        angles_deg = np.asarray([0,90.0,180.0], dtype='float32')
        angles_rad = angles_deg * np.pi /180.0
        source_position = [0,-2,0]
        detector_position = [0,1,0]

        self.ag = AcquisitionGeometry.create_Cone3D(source_position, detector_position)\
                                      .set_angles(angles_rad, angle_unit='radian')\
                                      .set_labels(['vertical', 'angle','horizontal'])\
                                      .set_panel((pixels_x,pixels_y), (0.1,0.2))
        
        self.ag_deg = AcquisitionGeometry.create_Cone3D(source_position, detector_position)\
                                      .set_angles(angles_deg, angle_unit='degree')\
                                      .set_labels(['vertical', 'angle','horizontal'])\
                                      .set_panel((pixels_x,pixels_y), (0.1,0.2))
        
        self.ig = self.ag.get_ImageGeometry()

        source_position_set = []
        detector_position_set = []
        detector_direction_x_set = []
        detector_direction_y_set = []

        for angle in angles_rad:
            rotation_matrix = np.eye(3)
            rotation_matrix[0,0] = rotation_matrix[1,1] = math.cos(-angle)
            rotation_matrix[0,1] = -math.sin(-angle)
            rotation_matrix[1,0] = math.sin(-angle)

            source_position_set.append(rotation_matrix.dot(source_position))
            detector_position_set.append(rotation_matrix.dot(detector_position))
            detector_direction_x_set.append(rotation_matrix.dot([1,0,0]))
            detector_direction_y_set.append(rotation_matrix.dot([0,0,1]))


        self.ag_Flex = AcquisitionGeometry.create_Cone3D_Flex(source_position_set=source_position_set,\
                                                         detector_position_set=detector_position_set,\
                                                         detector_direction_x_set=detector_direction_x_set,\
                                                         detector_direction_y_set=detector_direction_y_set)\
                                      .set_labels(['vertical', 'projection','horizontal'])\
                                      .set_panel((pixels_x,pixels_y), (0.1,0.2))
        

    @unittest.skipUnless(has_astra, "Requires ASTRA")
    def test_convert_geometry_simple(self):
        astra_vol, astra_sino = convert_geometry_to_astra(self.ig, self.ag)
        self.assertEqual(astra_sino['type'], 'cone')
        self.assertEqual(astra_sino['DistanceOriginSource'], self.ag.dist_source_center)
        self.assertEqual(astra_sino['DistanceOriginDetector'], self.ag.dist_center_detector)
        self.assertEqual(astra_sino['DetectorColCount'], self.ag.pixel_num_h)
        self.assertEqual(astra_sino['DetectorRowCount'], self.ag.pixel_num_v)
        self.assertEqual(astra_sino['DetectorSpacingX'], self.ag.pixel_size_h)
        self.assertEqual(astra_sino['DetectorSpacingY'], self.ag.pixel_size_v)
        np.testing.assert_allclose(astra_sino['ProjectionAngles'], -self.ag.angles)

        # degrees
        astra_vol, astra_sino = convert_geometry_to_astra(self.ig, self.ag_deg)
        np.testing.assert_allclose(astra_sino['ProjectionAngles'], -self.ag.angles)

        # check the image geometry
        self.assertEqual(astra_vol['GridColCount'], self.ig.voxel_num_x)
        self.assertEqual(astra_vol['GridRowCount'], self.ig.voxel_num_y)
        self.assertEqual(astra_vol['GridSliceCount'], self.ig.voxel_num_z)

        self.assertEqual(astra_vol['option']['WindowMinX'], -self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxX'], self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMinY'], -self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxY'], self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)
        self.assertEqual(astra_vol['option']['WindowMinZ'], -self.ig.voxel_num_z * self.ig.voxel_size_z * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxZ'], self.ig.voxel_num_z * self.ig.voxel_size_z * 0.5)


    @unittest.skipUnless(has_astra, "Requires ASTRA")
    def test_convert_geometry_vector2D(self):

        astra_vol, astra_sino = convert_geometry_to_astra_vec_2D(self.ig, self.ag)
        self.assertEqual(astra_sino['type'], 'fanflat_vec')

        # note the sign of the y values differs from the astra 3D projector case
        vectors = np.zeros((3,6),dtype='float64')
        vectors[0][1] = 1 * self.ag.dist_source_center
        vectors[0][3] = -self.ag.dist_center_detector
        vectors[0][4] = self.ag.pixel_size_h

        vectors[1][0] = -1 * self.ag.dist_source_center
        vectors[1][2] = self.ag.dist_center_detector
        vectors[1][5] = 1 * self.ag.pixel_size_h

        vectors[2][1] = -self.ag.dist_source_center
        vectors[2][3] = 1 * self.ag.dist_center_detector
        vectors[2][4] = -1 * self.ag.pixel_size_h

        np.testing.assert_allclose(astra_sino['Vectors'], vectors, atol=1e-6)

        #degrees
        astra_vol, astra_sino = convert_geometry_to_astra_vec_2D(self.ig, self.ag_deg)
        np.testing.assert_allclose(astra_sino['Vectors'], vectors, atol=1e-6)

        # image geometry
        self.assertEqual(astra_vol['GridColCount'], self.ig.voxel_num_x)
        self.assertEqual(astra_vol['GridRowCount'], self.ig.voxel_num_y)

        self.assertEqual(astra_vol['option']['WindowMinX'], -self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxX'], self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMinY'], -self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxY'], self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)

    @unittest.skipUnless(has_astra, "Requires ASTRA")
    def test_convert_geometry_vector3D(self):

        astra_vol, astra_sino = convert_geometry_to_astra_vec_3D(self.ig, self.ag)
        self.assertEqual(astra_sino['type'], 'cone_vec')

        vectors = np.zeros((3,12),dtype='float64')
        vectors[0][1] = -1 * self.ag.dist_source_center
        vectors[0][4] = self.ag.dist_center_detector
        vectors[0][6] = self.ag.pixel_size_h
        vectors[0][11] = self.ag.pixel_size_v

        vectors[1][0] = -1 * self.ag.dist_source_center
        vectors[1][3] = self.ag.dist_center_detector
        vectors[1][7] = -1 * self.ag.pixel_size_h
        vectors[1][11] = self.ag.pixel_size_v

        vectors[2][1] = self.ag.dist_source_center
        vectors[2][4] = -1 * self.ag.dist_center_detector
        vectors[2][6] = -1 * self.ag.pixel_size_h
        vectors[2][11] = self.ag.pixel_size_v

        np.testing.assert_allclose(astra_sino['Vectors'], vectors, atol=1e-6)

        #degrees
        astra_vol, astra_sino = convert_geometry_to_astra_vec_3D(self.ig, self.ag_deg)
        np.testing.assert_allclose(astra_sino['Vectors'], vectors, atol=1e-6)

        # image geometry
        self.assertEqual(astra_vol['GridColCount'], self.ig.voxel_num_x)
        self.assertEqual(astra_vol['GridRowCount'], self.ig.voxel_num_y)
        self.assertEqual(astra_vol['GridSliceCount'], self.ig.voxel_num_z)

        self.assertEqual(astra_vol['option']['WindowMinX'], -self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxX'], self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMinY'], -self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxY'], self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)
        self.assertEqual(astra_vol['option']['WindowMinZ'], -self.ig.voxel_num_z * self.ig.voxel_size_z * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxZ'], self.ig.voxel_num_z * self.ig.voxel_size_z * 0.5)


    @unittest.skipUnless(has_astra, "Requires ASTRA")
    def test_convert_geometry_Flex(self):
        """
        Checks the Flex convention agrees with the standard cone geometry 3D
        """

        astra_vol, astra_sino = convert_geometry_to_astra_vec_3D(self.ig, self.ag_Flex)
        self.assertEqual(astra_sino['type'], 'cone_vec')

        vectors = np.zeros((3,12),dtype='float64')
        vectors[0][1] = -1 * self.ag.dist_source_center
        vectors[0][4] = self.ag.dist_center_detector
        vectors[0][6] = self.ag.pixel_size_h
        vectors[0][11] = self.ag.pixel_size_v

        vectors[1][0] = -1 * self.ag.dist_source_center
        vectors[1][3] = self.ag.dist_center_detector
        vectors[1][7] = -1 * self.ag.pixel_size_h
        vectors[1][11] = self.ag.pixel_size_v

        vectors[2][1] = self.ag.dist_source_center
        vectors[2][4] = -1 * self.ag.dist_center_detector
        vectors[2][6] = -1 * self.ag.pixel_size_h
        vectors[2][11] = self.ag.pixel_size_v

        np.testing.assert_allclose(astra_sino['Vectors'], vectors, atol=1e-6)

        # image geometry
        self.assertEqual(astra_vol['GridColCount'], self.ig.voxel_num_x)
        self.assertEqual(astra_vol['GridRowCount'], self.ig.voxel_num_y)
        self.assertEqual(astra_vol['GridSliceCount'], self.ig.voxel_num_z)

        self.assertEqual(astra_vol['option']['WindowMinX'], -self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxX'], self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5)
        self.assertEqual(astra_vol['option']['WindowMinY'], -self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxY'], self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5)
        self.assertEqual(astra_vol['option']['WindowMinZ'], -self.ig.voxel_num_z * self.ig.voxel_size_z * 0.5)
        self.assertEqual(astra_vol['option']['WindowMaxZ'], self.ig.voxel_num_z * self.ig.voxel_size_z * 0.5)



class TestGeometry_Cone3D_Flex(unittest.TestCase):
    def setUp(self):
        self.pixels_x = 128
        self.pixels_y = 3

        # generate a set of source and detector positions
        self.source_position_set = [
            [60.39,	90.29,	63.12],
            [34.64,	24.1,	-66.48],
            [1.24,	66.74,	-98.15],
        ]

        self.detector_position_set = [
            [6579.59,	6069.16,	-9907.19],
            [4984.46,	5533.88,	1186.63],
            [-7528.65,	2598.33,	-7634.18],
        ]

        # detector direction x and y must be orthogonal
        self.detector_direction_x_set = [
            [1.2,	7.91,	-4.95],
            [4.66,	6.75,	-0.77],
            [-7.12,	4.69,	0.59]
        ]

        vec = [
            [-4.97,	5.62,	-8.26],
            [-0.97,	-6.93,	7.65],
            [-5.5,	7.43,	-5.14]
        ]

        self.detector_direction_y_set = []
        for i in range(3):
            self.detector_direction_x_set[i] = self.detector_direction_x_set[i] / np.linalg.norm(self.detector_direction_x_set[i])
            detector_direction_y = np.cross(self.detector_direction_x_set[i], vec[i])
            self.detector_direction_y_set.append(detector_direction_y / np.linalg.norm(detector_direction_y))

        self.ag = AcquisitionGeometry.create_Cone3D_Flex(source_position_set=self.source_position_set,\
                                                         detector_position_set=self.detector_position_set,\
                                                         detector_direction_x_set=self.detector_direction_x_set,\
                                                         detector_direction_y_set=self.detector_direction_y_set) \
                                      .set_labels(['vertical', 'projection','horizontal'])\
                                      .set_panel((self.pixels_x, self.pixels_y), (0.1,0.2))
        
        self.ig = create_cone_flex_default_ig(self.ag)

  

    @unittest.skipUnless(has_astra, "Requires ASTRA")
    def test_convert_geometry_simple(self):

        with self.assertRaises(ValueError):
            astra_vol, astra_sino = convert_geometry_to_astra(self.ig, self.ag)

    @unittest.skipUnless(has_astra, "Requires ASTRA")
    def test_convert_geometry_vector2D(self):
        with self.assertRaises(ValueError):
            astra_vol, astra_sino = convert_geometry_to_astra_vec_2D(self.ig, self.ag)

    @unittest.skipUnless(has_astra, "Requires ASTRA")
    def test_convert_geometry_vector3D(self):

        astra_vol, astra_sino = convert_geometry_to_astra_vec_3D(self.ig, self.ag)
        self.assertEqual(astra_sino['type'], 'cone_vec')

        for i in range(3):
            np.testing.assert_allclose(astra_sino['Vectors'][i][0:3], self.source_position_set[i], atol=1e-6)
            np.testing.assert_allclose(astra_sino['Vectors'][i][3:6], self.detector_position_set[i], atol=1e-6)
            np.testing.assert_allclose(astra_sino['Vectors'][i][6:9], self.detector_direction_x_set[i] * self.ag.pixel_size_h, atol=1e-6)
            np.testing.assert_allclose(astra_sino['Vectors'][i][9:12], self.detector_direction_y_set[i] * self.ag.pixel_size_v, atol=1e-6)
            
        # image geometry
        self.assertEqual(astra_vol['GridColCount'], self.ig.voxel_num_x)
        self.assertEqual(astra_vol['GridRowCount'], self.ig.voxel_num_y)
        self.assertEqual(astra_vol['GridSliceCount'], self.ig.voxel_num_z)

        origin = self.ag.config.system.volume_centre.position
        self.assertEqual(astra_vol['option']['WindowMinX'], -self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5 + origin[0])
        self.assertEqual(astra_vol['option']['WindowMaxX'], self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5 + origin[0])
        self.assertEqual(astra_vol['option']['WindowMinY'], -self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5 + origin[1])
        self.assertEqual(astra_vol['option']['WindowMaxY'], self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5 + origin[1])
        self.assertEqual(astra_vol['option']['WindowMinZ'], -self.ig.voxel_num_z * self.ig.voxel_size_z * 0.5 + origin[2])
        self.assertEqual(astra_vol['option']['WindowMaxZ'], self.ig.voxel_num_z * self.ig.voxel_size_z * 0.5 + origin[2])


        # with manual centre
        origin = self.ag.config.system.volume_centre.position = [1,-2,3.5]
        astra_vol, astra_sino = convert_geometry_to_astra_vec_3D(self.ig, self.ag)

        self.assertEqual(astra_vol['option']['WindowMinX'], -self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5 + origin[0])
        self.assertEqual(astra_vol['option']['WindowMaxX'], self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5 + origin[0])
        self.assertEqual(astra_vol['option']['WindowMinY'], -self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5 + origin[1])
        self.assertEqual(astra_vol['option']['WindowMaxY'], self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5 + origin[1])
        self.assertEqual(astra_vol['option']['WindowMinZ'], -self.ig.voxel_num_z * self.ig.voxel_size_z * 0.5 + origin[2])
        self.assertEqual(astra_vol['option']['WindowMaxZ'], self.ig.voxel_num_z * self.ig.voxel_size_z * 0.5 + origin[2])


    @unittest.skipUnless(has_astra, "Requires ASTRA")
    def test_convert_geometry_vector_data_origin(self):

        self.ag.set_panel((self.pixels_x, self.pixels_y), (0.1,0.2),origin="top-right")
        astra_vol, astra_sino = convert_geometry_to_astra_vec_3D(self.ig, self.ag)
        self.assertEqual(astra_sino['type'], 'cone_vec')

        for i in range(3):
            np.testing.assert_allclose(astra_sino['Vectors'][i][0:3], self.source_position_set[i], atol=1e-6)
            np.testing.assert_allclose(astra_sino['Vectors'][i][3:6], self.detector_position_set[i], atol=1e-6)
            np.testing.assert_allclose(astra_sino['Vectors'][i][6:9], -self.detector_direction_x_set[i] * self.ag.pixel_size_h, atol=1e-6)
            np.testing.assert_allclose(astra_sino['Vectors'][i][9:12], -self.detector_direction_y_set[i] * self.ag.pixel_size_v, atol=1e-6)
            
        # image geometry
        self.assertEqual(astra_vol['GridColCount'], self.ig.voxel_num_x)
        self.assertEqual(astra_vol['GridRowCount'], self.ig.voxel_num_y)
        self.assertEqual(astra_vol['GridSliceCount'], self.ig.voxel_num_z)

        origin = self.ag.config.system.volume_centre.position
        self.assertEqual(astra_vol['option']['WindowMinX'], -self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5 + origin[0])
        self.assertEqual(astra_vol['option']['WindowMaxX'], self.ig.voxel_num_x * self.ig.voxel_size_x * 0.5 + origin[0])
        self.assertEqual(astra_vol['option']['WindowMinY'], -self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5 + origin[1])
        self.assertEqual(astra_vol['option']['WindowMaxY'], self.ig.voxel_num_y * self.ig.voxel_size_y * 0.5 + origin[1])
        self.assertEqual(astra_vol['option']['WindowMinZ'], -self.ig.voxel_num_z * self.ig.voxel_size_z * 0.5 + origin[2])
        self.assertEqual(astra_vol['option']['WindowMaxZ'], self.ig.voxel_num_z * self.ig.voxel_size_z * 0.5 + origin[2])




#  Copyright 2020 United Kingdom Research and Innovation
#  Copyright 2020 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

from cil.framework import Processor, AcquisitionData
from cil.framework.labels import AcquisitionType
import numpy as np

import logging

log = logging.getLogger(__name__)

class CofR_xcorrelation(Processor):

    r'''CofR_xcorrelation processor uses the cross-correlation algorithm on a single slice between two projections at 180 degrees inteval.

    For use on parallel-beam geometry it requires two projections 180 degree apart.

    Parameters
    ----------
    slice_index: int or str, optional
        An integer defining the vertical slice to run the algorithm on or string='centre' specifying the central slice should be used (default is 'centre')
    projection_index: int or list/tuple of ints, optional
        The index of the first projection the cross correlation algorithm will use, where the second projection is chosen as the projection closest to 180 degrees from this.
        Or a list/tuple of ints specifying the two indices to be used for cross correlation (default is 0)
    ang_tol: float, optional
        The angular tolerance in degrees between the two input projections 180 degree gap (default is 0.1)

    Returns
    -------
    AcquisitionData
        object with an AcquisitionGeometry with updated centre of rotation

    '''

    def __init__(self, slice_index='centre', projection_index=0, ang_tol=0.1):

        kwargs = {
                    'slice_index': slice_index,
                    'ang_tol': ang_tol,
                    'projection_index': projection_index,
                    '_indices' : None,
                 }

        super(CofR_xcorrelation, self).__init__(**kwargs)

    def check_input(self, data):
        if not isinstance(data, AcquisitionData):
            raise Exception('Processor supports only AcquisitionData')

        if data.geometry == None:
            raise Exception('Geometry is not defined.')

        if data.geometry.geom_type == 'cone' or (data.geometry.geom_type & AcquisitionType.CONE_FLEX):
            raise ValueError("Only parallel-beam data is supported with this algorithm")

        if data.geometry.channels > 1:
            raise ValueError("Only single channel data is supported with this algorithm")

        if self.slice_index != 'centre':
            try:
                int(self.slice_index)
            except:
                raise ValueError("slice_index expected to be a positive integer or the string 'centre'. Got {0}".format(self.slice_index))

            if self.slice_index < 0 or self.slice_index >= data.get_dimension_size('vertical'):
                raise ValueError('slice_index is out of range. Must be in range 0-{0}. Got {1}'.format(data.get_dimension_size('vertical'), self.slice_index))

        # check if projection_index is either a tuple or list of length 2
        try:
            len_check = ( len(self.projection_index) == 2 )
            index = list(self.projection_index)
        # if projection_index does not have a length, put the object in a list
        except:
            len_check = True
            index = [self.projection_index]

        if not len_check:
            raise ValueError('Expected projection_index to be an int or list/tuple of 2 ints, got {0}'.format(self.projection_index))

        for angle in index:
            # check if all the indices are int
            if not isinstance(angle, int):
                raise ValueError('Expected projection_index to be an int, got {0}'.format(type(angle)))

            # check if all the indices are in range 0 to the number of angles
            if angle< 0  or angle>=data.geometry.config.angles.num_positions:
                raise ValueError('projection_index is out of range. Must be between 0 and {0}. Got {1}'.format(data.geometry.config.angles.num_positions-1, self.projection_index))

        angles_deg = data.geometry.config.angles.angle_data.copy()
        if data.geometry.config.angles.angle_unit == "radian":
                angles_deg *= 180/np.pi

        # if there is only 1 index specified, find the angle in the list closest to 180 degrees away from the index
        if len(index) == 1:
            new_index = CofR_xcorrelation._return_180_index(angles_deg, self.projection_index)
            index.append(new_index)

        # check the two angles are 180 degrees apart within the specified tolerance
        ang_diff = (angles_deg[index[0]] - angles_deg[index[1]]) % 360
        if abs(ang_diff-180) > self.ang_tol:
            if isinstance(self.projection_index, (list, tuple)):
                raise ValueError('Method requires projections 180+/-{0} degrees apart. The projection angle {1} and angle {2} provided are not within this tolerance.\
                            \nPick different projections or increase the angular tolerance ang_tol.'.format(self.ang_tol, data.geometry.angles[index[0]], data.geometry.angles[index[1]]))
            else:
                raise ValueError('Method requires projections 180+/-{0} degrees apart, for chosen projection angle {1} found closest angle {2}.\
                            \nPick a different initial projection or increase the angular tolerance `ang_tol`.'.format(self.ang_tol, data.geometry.angles[index[0]], data.geometry.angles[index[1]]))

        self._indices = index

        return True

    @staticmethod
    def _return_180_index(angles_deg, initial_index):
        '''
        Finds the index of the angle closest to 180 degrees from a specified initial angle, from a list of angles in degrees

        '''
        # set the target projection to 0
        angles_deg -= angles_deg[initial_index]

        #keep angles in range 0 to 360
        angles_deg = angles_deg % 360

        return np.abs(angles_deg - 180).argmin()

    def process(self, out=None):
        data = data_full = self.get_input()
        if AcquisitionType.DIM3 & data_full.geometry.dimension:
            data = data.get_slice(vertical=self.slice_index)

        geometry = data.geometry

        #cross correlate single slice with the 180deg one reversed
        data1 = data.get_slice(angle=self._indices[0]).as_array()
        data2 = np.flip(data.get_slice(angle=self._indices[1]).as_array())

        border = int(data1.size * 0.05)
        lag = np.correlate(data1[border:-border],data2[border:-border],"full")

        ind = lag.argmax()

        #fit quadratic to 3 centre points
        a = (lag[ind+1] + lag[ind-1] - 2*lag[ind]) * 0.5
        b = a + lag[ind] - lag[ind-1]
        quad_max = -b / (2*a) + ind

        shift = (quad_max - (lag.size-1)/2)/2
        shift = np.floor(shift *100 +0.5)/100

        new_geometry = data_full.geometry.copy()

        #set up new geometry
        new_geometry.config.system.rotation_axis.position[0] = shift * geometry.config.panel.pixel_size[0]

        log.info("Centre of rotation correction found using cross-correlation")
        log.info("Calculated from slice: %s", str(self.slice_index))
        log.info("Centre of rotation shift = %f pixels", shift)
        log.info("Centre of rotation shift = %f units at the object", shift * geometry.config.panel.pixel_size[0])
        log.info("Return new dataset with centred geometry")

        if out is None:
            return AcquisitionData(array = data_full, deep_copy = True, geometry = new_geometry)
        else:
            out.geometry = new_geometry
            return out



#  Copyright 2024 United Kingdom Research and Innovation
#  Copyright 2024 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# - CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

from abc import ABC, abstractmethod
import numpy
from numbers import Number
import logging

log = logging.getLogger(__name__)

class StepSizeRule(ABC):
    """
    Abstract base class for a step size rule. The abstract method, `get_step_size` takes in an algorithm and thus can access all parts of the algorithm (e.g. current iterate, current gradient, objective functions etc) and from this  should return a float as a step size. 
    """

    def __init__(self):
        '''Initialises the step size rule 
        '''
        pass

    @abstractmethod
    def get_step_size(self, algorithm):
        """
        Returns
        --------
        the calculated step size:float 
        """
        pass


class ConstantStepSize(StepSizeRule):
    """
    Step-size rule that always returns a constant step-size. 

    Parameters
    ----------
    step_size: float
        The step-size to be returned with each call. 
    """

    def __init__(self, step_size):
        '''Initialises the constant step size rule
        
         Parameters:
         -------------
         step_size : float, the constant step size 
        '''
        self.step_size = step_size

    def get_step_size(self, algorithm):
        """
        Returns
        --------
        the calculated step size:float
        """
        return self.step_size


class ArmijoStepSizeRule(StepSizeRule):

    r""" Applies the Armijo rule to calculate the step size (step_size).

    The Armijo rule runs a while loop to find the appropriate step_size by starting from a very large number (`alpha`). The step_size is found by reducing the step size (by a factor `beta`) in an iterative way until a certain criterion is met. To avoid infinite loops, we add a maximum number of times (`max_iterations`) the while loop is run.

    Reference
    ---------
    - Algorithm 3.1 in Nocedal, J. and Wright, S.J. eds., 1999. Numerical optimization. New York, NY: Springer New York. https://www.math.uci.edu/~qnie/Publications/NumericalOptimization.pdf)
    
    - https://projecteuclid.org/download/pdf_1/euclid.pjm/1102995080
    
    
    Parameters
    ----------
    alpha: float, optional, default=1e6
        The starting point for the step size iterations 
    beta: float between 0 and 1, optional, default=0.5
        The amount the step_size is reduced if the criterion is not met
    max_iterations: integer, optional, default is numpy.ceil (2 * numpy.log10(alpha) / numpy.log10(2))
        The maximum number of iterations to find a suitable step size 
    warmstart: Boolean, default is True
        If `warmstart = True` the initial step size at each Armijo iteration is the calculated step size from the last iteration. If `warmstart = False` at each  Armijo iteration, the initial step size is reset to the original, large `alpha`. 
        In the case of *well-behaved* convex functions, `warmstart = True` is likely to be computationally less expensive. In the case of non-convex functions, or particularly tricky functions, setting `warmstart = False` may be beneficial. 

    """

    def __init__(self, alpha=1e6, beta=0.5, max_iterations=None, warmstart=True):
        '''Initialises the step size rule 
        '''
        
        self.alpha_orig = alpha
        if self.alpha_orig is None: # Can be removed when alpha and beta are deprecated in GD
            self.alpha_orig = 1e6 
        self.alpha = self.alpha_orig
        self.beta = beta 
        if self.beta is None:  # Can be removed when alpha and beta are deprecated in GD
            self.beta = 0.5
            
        self.max_iterations = max_iterations
        if self.max_iterations is None:
            self.max_iterations = numpy.ceil(2 * numpy.log10(self.alpha_orig) / numpy.log10(2))
            
        self.warmstart=warmstart

    def get_step_size(self, algorithm):
        """
        Applies the Armijo rule to calculate the step size (`step_size`)

        Returns
        --------
        the calculated step size:float

        """
        k = 0
        if not self.warmstart:  
            self.alpha = self.alpha_orig
        
        f_x = algorithm.calculate_objective_function_at_point(algorithm.solution)

        self.x_armijo = algorithm.solution.copy()
        
        log.debug("Starting Armijo backtracking with initial step size: %f", self.alpha)
        
        while k < self.max_iterations:

            algorithm.gradient_update.multiply(self.alpha, out=self.x_armijo)
            algorithm.solution.subtract(self.x_armijo, out=self.x_armijo)

            f_x_a = algorithm.calculate_objective_function_at_point(self.x_armijo)
            sqnorm = algorithm.gradient_update.squared_norm()
            if f_x_a - f_x <= - (self.alpha/2.) * sqnorm:
                break
            k += 1.
            self.alpha *= self.beta
        
        log.info("Armijo rule took %d iterations to find step size", k)

        if k == self.max_iterations:
            raise ValueError(
                'Could not find a proper step_size in {} loops. Consider increasing alpha or max_iterations.'.format(self.max_iterations))
        
        return self.alpha


class BarzilaiBorweinStepSizeRule(StepSizeRule):

    r""" Applies the Barzilai- Borwein rule to calculate the step size (step_size).

    Let :math:`\Delta x=x_k-x_{k-1}` and :math:`\Delta g=g_k-g_{k-1}`. Where :math:`x_k` is the :math:`k` th iterate (current solution after iteration :math:`k` ) and :math:`g_k` is the gradient calculation in the :math:`k` th iterate, found in :code:`algorithm.gradient_update`.  A Barzilai-Borwein (BB) iteration is :math:`x_{k+1}=x_k-\alpha_kg_k` where the step size :math:`\alpha _k` is either

    - :math:`\alpha_k^{LONG}=\frac{\Delta x\cdot\Delta x}{\Delta x\cdot\Delta g}`, or

    - :math:`\alpha_k^{SHORT}=\frac{\Delta x \cdot\Delta g}{\Delta g \cdot\Delta g}`.
    
    Where the operator :math:`\cdot` is the standard inner product between two vectors. 
    
    This is suitable for use with gradient based iterative methods where the calculated gradient is stored as `algorithm.gradient_update`.
    
    Parameters
    ----------
    initial: float, greater than zero 
        The step-size for the first iteration. We recommend something of the order :math:`1/f.L` where :math:`f` is the (differentiable part of) the objective you wish to minimise.
    mode: One of 'long', 'short' or 'alternate', default is 'short'. 
        This calculates the step-size based on the LONG, SHORT or alternating between the two, starting with short. 
    stabilisation_param: 'auto', float or 'off', default is 'auto'
        In order to add stability the step-size has an upper limit of :math:`\Delta/\|g_k\|` where by 'default', the `stabilisation_param`, :math:`\Delta` is  determined automatically to be the minimium of :math:`\Delta x` from the first 3 iterations. The user can also pass a fixed constant or turn "off" the stabilisation, equivalently passing `np.inf`.
        
    
    Reference
    ---------
    - Barzilai, Jonathan; Borwein, Jonathan M. (1988). "Two-Point Step Size Gradient Methods". IMA Journal of Numerical Analysis. 8: 141–148, https://doi.org/10.1093/imanum/8.1.141
    
    - Burdakov, O., Dai, Y. and Huang, N., 2019. STABILIZED BARZILAI-BORWEIN METHOD. Journal of Computational Mathematics, 37(6). https://doi.org/10.4208/jcm.1911-m2019-0171

    - https://en.wikipedia.org/wiki/Barzilai-Borwein_method
    """

    def __init__(self, initial, mode='short', stabilisation_param="auto"):
        '''Initialises the step size rule 
        '''
 
        self.mode=mode
        if self.mode == 'short':
            self.is_short = True
        elif self.mode == 'long' or self.mode == 'alternate':
            self.is_short = False
        else:
            raise ValueError('Mode should be chosen from "long", "short" or "alternate". ')
        
        self.store_grad=None 
        self.store_x=None
        self.initial=initial
        if stabilisation_param == 'auto':
            self.adaptive = True
            stabilisation_param = numpy.inf
        elif stabilisation_param == "off":
            self.adaptive = False 
            stabilisation_param = numpy.inf
        elif ( isinstance(stabilisation_param, Number) and stabilisation_param >=0):
            self.adaptive = False 
        else:
            raise TypeError(" The stabilisation_param should be 'auto', a positive number or 'off'")
        self.stabilisation_param=stabilisation_param
        
    

    def get_step_size(self, algorithm):
        """
        Applies the B-B rule to calculate the step size (`step_size`)

        Returns
        --------
        the calculated step size:float

        """
        #For the first iteration we use an initial step size because the BB step size requires a previous iterate. 
        if self.store_x is None:
            self.store_x=algorithm.x.copy() # We store the last iterate in order to calculate the BB step size 
            self.store_grad=algorithm.gradient_update.copy()# We store the last gradient in order to calculate the BB step size 
            return self.initial
        
        gradient_norm = algorithm.gradient_update.norm()
        #If the gradient is zero, gradient based algorithms will not update and te step size calculation will divide by zero so we stop iterations. 
        if gradient_norm < 1e-8:
            raise StopIteration

        algorithm.x.subtract(self.store_x, out=self.store_x) 
        algorithm.gradient_update.subtract(self.store_grad, out=self.store_grad)
        if self.is_short:
                ret = (self.store_x.dot(self.store_grad))/ (self.store_grad.dot(self.store_grad))
        else:
            ret = (self.store_x.dot(self.store_x))/ (self.store_x.dot(self.store_grad))

        
        #This computes the default stabilisation parameter, using the first three iterations
        if (algorithm.iteration <=3 and self.adaptive):
            self.stabilisation_param = min(self.stabilisation_param, self.store_x.norm() )
        
        # Computes the step size as the minimum of the ret, above, and :math:`\Delta/\|g_k\|` ignoring any NaN values. 
        ret = numpy.nanmin( numpy.array([ret, self.stabilisation_param/gradient_norm]))
        
        # We store the last iterate and gradient in order to calculate the BB step size 
        self.store_x.fill(algorithm.x)
        self.store_grad.fill(algorithm.gradient_update)
        
        if self.mode == "alternate":
            self.is_short =  not self.is_short       
        
        return ret


#  Copyright 2020 United Kingdom Research and Innovation
#  Copyright 2020 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

from cil.optimisation.algorithms import Algorithm
import logging

log = logging.getLogger(__name__)


class LADMM(Algorithm):
    r"""
    LADMM is the Linearized Alternating Direction Method of Multipliers (LADMM).

    The general form of ADMM is given by the following optimization problem: 
    
    .. math::
    
        \min_{x} f(x) + g(y), \text{ subject to } Ax + By = b

    In CIL, we have implemented the case where :math:`A = Id`, :math:`B = -K`, :math:`b = 0`  which gives 
    
    .. math::
        
        \min_x f(Kx) + g(x).
        
    The algorithm is given by the following iteration, for :math:`k\geq 1`:
    
    .. math::

        \begin{cases}
            x_{k} = \mathrm{prox}_{\tau f} \left(x_{k-1} - \dfrac{\tau}{\sigma} A_{T}\left(Ax_{k-1} - z_{k-1} + u_{k-1} \right)  \right)\\
            z_{k} = \mathrm{prox}_{\sigma g} \left(Ax_{k} + u_{k-1}\right) \\
            u_{k} = u_{k-1} + Ax_{k} - z_{k}
        \end{cases}
        
    where :math:`\mathrm{prox}_{\tau f}` is the proximal operator of :math:`f` and :math:`\mathrm{prox}_{\sigma g}` is the proximal operator of :math:`g`.
    
    
    Parameters
    ------------
    operator:  CIL Linear Operator
        Operator :math:`K` in the objective function
    f: CIL Function
        Convex function with "simple" proximal
    g: CIL Function
        Convex function with "simple" proximal
    sigma: float, positive, defaults to 1.
        Positive step size parameter
    tau: float, positive, defaults to :math:`\frac{\sigma}{\|K\|^{2}}`
        Positive step size parameter
    initial: DataContainer, defaults to DataContainer filled with zeros
        Initial guess 
            
    
    Note
    ----
    This implementation of ADMM minimises the same objective function as the Primal-Dual Hybrid Gradient (PDHG) method.
    The main algorithmic difference is that in ADMM we compute the proximal of :math:`f` and :math:`g` 
    where in the PDHG this is a proximal-conjugate and proximal.
    
    
    Note
    -----
    Reference (Section 8) : O’Connor, D., Vandenberghe, L. On the equivalence of the primal-dual hybrid gradient method and Douglas–Rachford splitting. Math. Program. 179, 85–108 (2020). https://doi.org/10.1007/s10107-018-1321-1

    """


    def __init__(self, f=None, g=None, operator=None, \
                       tau = None, sigma = 1.,
                       initial = None, **kwargs):

        """Initialisation of the algorithm."""

        super(LADMM, self).__init__(**kwargs)

        self.set_up(f = f, g = g, operator = operator, tau = tau,\
             sigma = sigma, initial=initial)

    def set_up(self, f, g, operator, tau = None, sigma=1., initial=None):
        """Set up of the algorithm."""
        log.info("%s setting up", self.__class__.__name__)

        self.f = f
        self.g = g
        self.operator = operator

        self.tau = tau
        self.sigma = sigma

        if self.tau is None:
            normK = self.operator.norm()
            self.tau = self.sigma / normK ** 2

        if initial is None:
            self.x = self.operator.domain_geometry().allocate(0)
        else:
            self.x = initial.copy()

        # allocate space for operator direct & adjoint
        self.tmp_dir = self.operator.range_geometry().allocate(0)
        self.tmp_adj = self.operator.domain_geometry().allocate(0)

        self.z = self.operator.range_geometry().allocate(0)
        self.u = self.operator.range_geometry().allocate(0)

        self.configured = True

        log.info("%s configured", self.__class__.__name__)

    def update(self):
        """Performs a single iteration of the LADMM algorithm"""
        self.tmp_dir += self.u
        self.tmp_dir -= self.z
        self.operator.adjoint(self.tmp_dir, out = self.tmp_adj)

        self.x.sapyb(1, self.tmp_adj, -(self.tau/self.sigma), out=self.x)

        # apply proximal of f
        tmp = self.f.proximal(self.x, self.tau)
        self.operator.direct(tmp, out=self.tmp_dir)
        # store the result in x
        self.x.fill(tmp)
        del tmp

        self.u += self.tmp_dir

        # apply proximal of g
        self.g.proximal(self.u, self.sigma, out = self.z)

        # update
        self.u -= self.z

    def update_objective(self):
        """Update the objective function value"""
        self.loss.append(self.f(self.x) +  self.g(self.operator.direct(self.x)) )



#  Copyright 2019 United Kingdom Research and Innovation
#  Copyright 2019 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

from cil.optimisation.algorithms import Algorithm
from cil.optimisation.functions import ZeroFunction
from cil.optimisation.utilities import ConstantStepSize, StepSizeRule
import numpy
import logging
from numbers import Real, Number
import warnings

log = logging.getLogger(__name__)


class ISTA(Algorithm):

    r"""Iterative Shrinkage-Thresholding Algorithm (ISTA), see :cite:`BeckTeboulle_b`, :cite:`BeckTeboulle_a`, is used to solve:

    .. math:: \min_{x} f(x) + g(x)

    where :math:`f` is differentiable, and :math:`g` has a *simple* proximal operator.

    In each update, the algorithm computes:

    .. math:: x_{k+1} = \mathrm{prox}_{\alpha g}(x_{k} - \alpha\nabla f(x_{k}))

    where :math:`\alpha` is the :code:`step_size`.

    Note
    ----

    For a constant step size, :math:`\alpha`, convergence of ISTA
    is guaranteed if

    .. math:: \alpha\in(0, \frac{2}{L}),

    where :math:`L` is the Lipschitz constant of :math:`f`, see :cite:`CombettesValerie`.

    Parameters
    ----------

    initial : DataContainer
              Initial guess of ISTA. :math:`x_{0}`
    f : Function
        Differentiable function. If `None` is passed, the algorithm will use the ZeroFunction.
    g : Function or `None`
        Convex function with *simple* proximal operator. If `None` is passed, the algorithm will use the ZeroFunction.
    step_size : positive :obj:`float` or child class of :meth:`cil.optimisation.utilities.StepSizeRule`',  default = None
                Step size for the gradient step of ISTA. If a float is passed, this is used as a constant step size.  If a child class of :meth:`cil.optimisation.utilities.StepSizeRule` is passed then its method :meth:`get_step_size` is called for each update. 
                The default :code:`step_size` is a constant :math:`\frac{0.99*2}{L}` or 1 if `f=None`.
    preconditioner: class with an `apply` method or a function that takes an initialised CIL function as an argument and modifies a provided `gradient`.
            This could be a custom `preconditioner` or one provided in :meth:`~cil.optimisation.utilities.preconditoner`. If None is passed then `self.gradient_update` will remain unmodified. 


    kwargs: Keyword arguments
        Arguments from the base class :class:`.Algorithm`.

    Note
    -----
    If the function `g` is set to `None` or to the `ZeroFunction` then the ISTA algorithm is equivalent to Gradient Descent.

    If the function `f` is set to `None` or to the `ZeroFunction` then the ISTA algorithm is equivalent to a Proximal Point Algorithm.

    Note
    ----
    The ISTA algorithm is equivalent to the Proximal Gradient Descent (PGD) algorithm. In CIL, the PGD algorithm is an alias for the ISTA algorithm.
    
    Examples
    --------

    .. math:: \underset{x}{\mathrm{argmin}}\|A x - b\|^{2}_{2}

    >>> f = LeastSquares(A, b=b, c=0.5)
    >>> g = ZeroFunction()
    >>> ig = Aop.domain
    >>> ista = ISTA(initial = ig.allocate(), f = f, g = g)
    >>> ista.run()


    See also
    --------

    :class:`.FISTA`
    :class:`.GD`


    """

    def _provable_convergence_condition(self):
        if self.preconditioner is not None:
            raise NotImplementedError(
                "Can't check convergence criterion if a preconditioner is used ")

        if isinstance(self.step_size_rule, ConstantStepSize):
            return self.step_size_rule.step_size <= 0.99*2.0/self.f.L
        else:
            raise TypeError(
                "Can't check convergence criterion for non-constant step size")

    @property
    def step_size(self):
        '''
        Returns the most recently used step size. Note, if the step-size is set by a non-constant step size rule, you must use the algorithm run or update method before this getter will return the most recently used step size. 
        '''
        
        if isinstance(self.step_size_rule, ConstantStepSize):
            return self.step_size_rule.step_size
        else:
            try: 
                return self._step_size
            except AttributeError:
                raise NotImplementedError("Note the step-size is set by a step-size rule and could change with each iteration. Call the algorithm run or update method first and then this function will give the most recently used step size.")

    # Set default step size
    def _calculate_default_step_size(self):
        """ Calculates the default step size if a step size rule or a step size is not provided. 
        """

        
        return 0.99*2.0/self.f.L

        

    def __init__(self, initial, f, g, step_size=None, preconditioner=None, **kwargs):

        super(ISTA, self).__init__(**kwargs)
        self.set_up(initial=initial, f=f, g=g, step_size=step_size,
                    preconditioner=preconditioner, **kwargs)

    def set_up(self, initial, f, g, step_size, preconditioner, **kwargs):
        """Set up of the algorithm"""
        log.info("%s setting up", self.__class__.__name__)
        # set up ISTA
        self.initial = initial
        self.x_old = initial.copy()
        self.x = initial.copy()
        self.gradient_update = initial.copy()

        if f is None:
            f = ZeroFunction()

        self.f = f

        if g is None:
            g = ZeroFunction()

        self.g = g

        if isinstance(f, ZeroFunction) and isinstance(g, ZeroFunction):
            raise ValueError(
                'You set both f and g to be the ZeroFunction and thus the iterative method will not update and will remain fixed at the initial value.')

        # set step_size
        if step_size is None:
            self.step_size_rule = ConstantStepSize(
                self._calculate_default_step_size())
        elif isinstance(step_size, Real):
            self.step_size_rule = ConstantStepSize(step_size)
        elif isinstance(step_size, StepSizeRule):
            self.step_size_rule = step_size
        
        self.preconditioner = preconditioner

        self.configured = True
        log.info("%s configured", self.__class__.__name__)

    def update(self):
        r"""Performs a single iteration of ISTA

        .. math:: x_{k+1} = \mathrm{prox}_{\alpha g}(x_{k} - \alpha\nabla f(x_{k}))

        """

        # gradient step
        self.f.gradient(self.x_old, out=self.gradient_update)
        if self.preconditioner is not None:
            self.preconditioner.apply(
                self, self.gradient_update, out=self.gradient_update)

        try:
            self._step_size = self.step_size_rule.get_step_size(self)
        except NameError:
            raise NameError(msg='`step_size` must be `None`, a real float or a child class of :meth:`cil.optimisation.utilities.StepSizeRule`')

        self.x_old.sapyb(1., self.gradient_update, -self._step_size, out=self.x_old)

        # proximal step
        self.g.proximal(self.x_old, self._step_size, out=self.x)

    def _update_previous_solution(self):
        """ Swaps the references to current and previous solution based on the :func:`~Algorithm.update_previous_solution` of the base class :class:`Algorithm`.
        """
        tmp = self.x_old
        self.x_old = self.x
        self.x = tmp

    def get_output(self):
        " Returns the current solution. "
        return self.x_old

    def update_objective(self):
        """ Updates the objective

        .. math:: f(x) + g(x)

        """
        self.loss.append(self.calculate_objective_function_at_point(self.x_old))

    def calculate_objective_function_at_point(self, x):
        """ Calculates the objective at a given point x

        .. math:: f(x) + g(x)
        
        Parameters
        ----------
        x : DataContainer
        
        """
        return self.f(x) + self.g(x)

class FISTA(ISTA):

    r"""Fast Iterative Shrinkage-Thresholding Algorithm (FISTA), see :cite:`BeckTeboulle_b`, :cite:`BeckTeboulle_a`, is used to solve:

    .. math:: \min_{x} f(x) + g(x)

    where :math:`f` is differentiable and :math:`g` has a *simple* proximal operator.


    In each update the algorithm completes the following steps:

    .. math::

        \begin{cases}
                x_{k} = \mathrm{prox}_{\alpha g}(y_{k} - \alpha\nabla f(y_{k}))\\
                t_{k+1} = \frac{1+\sqrt{1+ 4t_{k}^{2}}}{2}\\
                y_{k+1} = x_{k} + \frac{t_{k}-1}{t_{k+1}}(x_{k} - x_{k-1})
        \end{cases}

    where :math:`\alpha` is the :code:`step_size`.

    Note that the above applies for :math:`k\geq 1`. For :math:`k=0`, :math:`x_{0}` and :math:`y_{0}` are initialised to `initial`, and :math:`t_{1}=1`.


    Parameters
    ----------

    initial : DataContainer
            Starting point of the algorithm
    f : Function
        Differentiable function.  If `None` is passed, the algorithm will use the ZeroFunction.
    g : Function or `None`
        Convex function with *simple* proximal operator. If `None` is passed, the algorithm will use the ZeroFunction.
    step_size : positive :obj:`float` or child class of :meth:`cil.optimisation.utilities.StepSizeRule`',  default = None
                Step size for the gradient step of ISTA. If a float is passed, this is used as a constant step size. If a child class of :meth:`cil.optimisation.utilities.StepSizeRule` is passed then it's method :meth:`get_step_size` is called for each update. 
                The default :code:`step_size` is a constant :math:`\frac{1}{L}` or 1 if `f=None`.
    preconditioner : class with an `apply` method or a function that takes an initialised CIL function as an argument and modifies a provided `gradient`.
            This could be a custom `preconditioner` or one provided in :meth:`~cil.optimisation.utilities.preconditoner`. If None is passed  then `self.gradient_update` will remain unmodified. 

    kwargs: Keyword arguments
        Arguments from the base class :class:`.Algorithm`.

    Note
    -----
    If the function `g` is set to `None` or to the `ZeroFunction` then the FISTA algorithm is equivalent to Accelerated Gradient Descent by Nesterov (:cite:`nesterov2003introductory` algorithm 2.2.9).

    If the function `f` is set to `None` or to the `ZeroFunction` then the FISTA algorithm is equivalent to Guler's First Accelerated Proximal Point Method  (:cite:`guler1992new` sec 2).

    Note
    -----
    FISTA is sometimes also known as Accelerated Proximal Gradient Descent (APGD). Note that in CIL, 'APGD' with the default Nesterov momentum is equivalent to 'FISTA' but the 'APGD' algorithm gives more options for momentum. 


    Examples
    --------

    .. math:: \underset{x}{\mathrm{argmin}}\|A x - b\|^{2}_{2}


    >>> f = LeastSquares(A, b=b, c=0.5)
    >>> g = ZeroFunction()
    >>> ig = Aop.domain
    >>> fista = FISTA(initial = ig.allocate(), f = f, g = g)
    >>> fista.run()

    See also
    --------
    :class:`.FISTA`
    :class:`.GD`

    """

    def _calculate_default_step_size(self):
        """Calculate the default step size if a step size rule or step size is not provided 
        """
        return 1./self.f.L



    def _provable_convergence_condition(self):
        if self.preconditioner is not None:
            raise NotImplementedError(
                "Can't check convergence criterion if a preconditioner is used ")

        if isinstance(self.step_size_rule, ConstantStepSize):
            return self.step_size_rule.step_size <= 1./self.f.L
        else:
            raise TypeError(
                "Can't check convergence criterion for non-constant step size")

    def __init__(self, initial, f, g, step_size=None,  preconditioner=None, **kwargs):

        self.y = initial.copy()
        self.t = 1
        super(FISTA, self).__init__(initial=initial, f=f, g=g,
                                    step_size=step_size,  preconditioner=preconditioner, **kwargs)

    def update(self):
        r"""Performs a single iteration of FISTA. For :math:`k\geq 1`:

        .. math::

            \begin{cases}
                x_{k} = \mathrm{prox}_{\alpha g}(y_{k} - \alpha\nabla f(y_{k}))\\
                t_{k+1} = \frac{1+\sqrt{1+ 4t_{k}^{2}}}{2}\\
                y_{k+1} = x_{k} + \frac{t_{k}-1}{t_{k+1}}(x_{k} - x_{k-1})
            \end{cases}

        """

        self.t_old = self.t

        self.f.gradient(self.y, out=self.gradient_update)

        if self.preconditioner is not None:
            self.preconditioner.apply(
                self, self.gradient_update, out=self.gradient_update)

        self._step_size = self.step_size_rule.get_step_size(self)

        self.y.sapyb(1., self.gradient_update, -self._step_size, out=self.y)

        self.g.proximal(self.y, self._step_size, out=self.x)

        self.t = 0.5*(1 + numpy.sqrt(1 + 4*(self.t_old**2)))

        self.x.subtract(self.x_old, out=self.y)
        self.y.sapyb(((self.t_old-1)/self.t), self.x, 1.0, out=self.y)




#  Copyright 2018 United Kingdom Research and Innovation
#  Copyright 2018 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt
# Joshua DM Hellier (University of Manchester) [refactorer]
import numpy
import warnings

from .labels import AcquisitionDimension, Backend, AcquisitionType
from .data_container import DataContainer
from .partitioner import Partitioner


class AcquisitionData(DataContainer, Partitioner):
    """
    DataContainer for holding 2D or 3D sinogram
    
    Parameters
    ----------
    array : numpy.ndarray or DataContainer
        The data array. Default None creates an empty array of size determined by the geometry.
    deep_copy : bool, default is True
        If True, the array will be deep copied. If False, the array will be shallow copied.
    geometry : AcquisitionGeometry
        The geometry of the data. If the dtype of the array and geometry are different, the geometry dtype will be overridden.
    
    **kwargs:
        dtype : numpy.dtype
            Specify the data type of the AcquisitionData array, this is useful if you pass None to array and want to over-ride the dtype of the geometry. 
            If an array is passed, dtype must match the dtype of the array.

    """
    __container_priority__ = 1

    @property
    def geometry(self):
        return self._geometry

    @geometry.setter
    def geometry(self, val):
        self._geometry = val

    @property
    def dimension_labels(self):
        return self.geometry.dimension_labels

    @dimension_labels.setter
    def dimension_labels(self, val):
        if val is not None:
            raise ValueError("Unable to set the dimension_labels directly. Use geometry.set_labels() instead")

    def __init__(self,
                 array = None,
                 deep_copy=True,
                 geometry = None,
                 **kwargs):

        dtype = kwargs.pop('dtype', None)
        if dtype is not None and array is not None:
            if dtype != array.dtype:
                    raise TypeError('dtype must match the array dtype got {} expected {}'.format(dtype, array.dtype))

        if geometry is None:
            raise AttributeError("AcquisitionData requires a geometry")

        labels = kwargs.pop('dimension_labels', None)
        if labels is not None and labels != geometry.dimension_labels:
                raise ValueError("Deprecated: 'dimension_labels' cannot be set with 'allocate()'. Use 'geometry.set_labels()' to modify the geometry before using allocate.")

        if array is None:
            if dtype is None:
                dtype = geometry.dtype
            array = numpy.empty(geometry.shape, dtype)
    
        elif issubclass(type(array) , DataContainer):
            array = array.as_array()

        elif issubclass(type(array) , numpy.ndarray):
            # remove singleton dimensions
            array = numpy.squeeze(array)

        else:
            raise TypeError('array must be a CIL type DataContainer or numpy.ndarray got {}'.format(type(array)))

        if array.shape != geometry.shape:
            raise ValueError('Shape mismatch got {} expected {}'.format(array.shape, geometry.shape))

        super(AcquisitionData, self).__init__(array, deep_copy, geometry=geometry,**kwargs)

        if kwargs:
            warnings.warn(f"Unused keyword arguments: {kwargs}", stacklevel=2)

    def __eq__(self, other):
        '''
        Check if two AcquisitionData objects are equal. This is done by checking if the geometry, data and dtype are equal.
        Also, if the other object is a numpy.ndarray, it will check if the data and dtype are equal.
        
        Parameters
        ----------
        other: AcquisitionData or numpy.ndarray
            The object to compare with.
        
        Returns
        -------
        bool
            True if the two objects are equal, False otherwise.
        '''

        if isinstance(other, AcquisitionData):
            if numpy.array_equal(self.as_array(), other.as_array()) \
                and self.geometry == other.geometry \
                and self.dtype == other.dtype:
                return True 
        elif numpy.array_equal(self.as_array(), other) and self.dtype==other.dtype:
            return True
        else:
            return False
        
    def _get_slice(self, **kwargs):
        '''
        Functionality of get_slice
        '''

        force = kwargs.pop('force', False)
        try:
            geometry_new = self.geometry.get_slice(**kwargs)
        except ValueError:
            if force:
                geometry_new = None
            else:
                raise ValueError ("Unable to return slice of requested AcquisitionData. Use 'force=True' to return DataContainer instead.")

        #get new data
        #if vertical = 'centre' slice convert to index and subset, this will interpolate 2 rows to get the center slice value
        if kwargs.get('vertical') == 'centre':
            dim = self.geometry.dimension_labels.index('vertical')

            centre_slice_pos = (self.geometry.shape[dim]-1) / 2.
            ind0 = int(numpy.floor(centre_slice_pos))
            w2 = centre_slice_pos - ind0
            kwargs['vertical'] = ind0
            out = DataContainer.get_slice(self, **kwargs)

            if w2 > 0:
                kwargs['vertical'] = ind0 +  1
                out2 = DataContainer.get_slice(self, **kwargs)
                out = out * (1 - w2) + out2 * w2
        else:
            out = DataContainer.get_slice(self, **kwargs)

        if len(out.shape) == 1 or geometry_new is None:
            return out
        else:
            return AcquisitionData(out.array, deep_copy=False, geometry=geometry_new)
        
    
    def get_slice(self, *args, **kwargs):
        '''
        Returns a new AcquisitionData of a single slice in the requested direction.

        Parameters
        ----------
        channel: int, optional
            index on channel dimension to slice on. If None, does not slice on this dimension.
        angle/projection: int, optional
            index on angle or projection dimension to slice on. Dimension label depends on the geometry type:
            For CONE_FLEX geometry, use 'projection'.
            For all other geometries, use 'angle'.
        vertical: int, str, optional
            If int, index on vertical dimension to slice on. If str, can be 'centre' to return the slice at the center of the vertical dimension.
        horizontal: int, optional
            index on horizontal dimension to slice on. If None, does not slice on this dimension.
        force: bool, default False
            If True, will return a DataContainer instead of an AcquisitionData when the requested slice is not reconstructable.

        Returns
        -------
        AcquisitionData
            A new AcquisitionData object containing the requested slice.
        DataContainer
            If `force=True` and the slice is not reconstructable, a DataContainer will be returned instead.

        Notes
        -----
        If there is an even number of slices in the vertical dimension and 'centre' is requested, the slice returned will be the average of the two middle slices.
        '''

        for key in kwargs:
            if self.geometry.geom_type == AcquisitionType.CONE_FLEX and key == 'angle':
                raise TypeError(f"Cannot use 'angle' for Cone3D_Flex geometry. Use 'projection' instead.")
            elif self.geometry.geom_type != AcquisitionType.CONE_FLEX and key == 'projection':
                raise TypeError(f"Cannot use 'projection' for geometries that use angles. Use 'angle' instead.")
            elif key not in AcquisitionDimension and key != 'force':
                raise TypeError(f"'{key}' not in allowed labels {AcquisitionDimension}.")
            
        if args:
            warnings.warn("Positional arguments for get_slice are deprecated. Use keyword arguments instead.", DeprecationWarning, stacklevel=2)
            
            num_args = len(args)

            if num_args > 0:
                kwargs['channel'] = args[0]

            if num_args > 1:
                if self.geometry.geom_type & AcquisitionType.CONE_FLEX:
                    kwargs['projection'] = args[1]
                else:
                    kwargs['angle'] = args[1]

            if num_args > 2:
                kwargs['vertical'] = args[2]

            if num_args > 3:
                kwargs['horizontal'] = args[3]

            if num_args > 4:
                kwargs['force'] = args[4]

        return self._get_slice(**kwargs)

    def reorder(self, order):
        '''
        Reorders the data in memory as requested. This is an in-place operation.

        Parameters
        ----------
        order: list or str
            Ordered list of labels from self.dimension_labels, or string 'astra' or 'tigre'.
        '''
        if order in Backend:
            order = AcquisitionDimension.get_order_for_engine(order, self.geometry)

        super().reorder(order)



#  Copyright 2021 United Kingdom Research and Innovation
#  Copyright 2021 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

from cil.framework import Processor, AcquisitionData
from cil.framework.labels import AcquisitionDimension, AcquisitionType
import scipy
import numpy as np
import logging
import math
import importlib

log = logging.getLogger(__name__)

class CofR_image_sharpness(Processor):

    """This creates a CentreOfRotationCorrector processor.

    The processor will find the centre offset by maximising the sharpness of a reconstructed slice.

    Can be used on single slice parallel-beam, and centre slice cone beam geometry. For use only with datasets that can be reconstructed with FBP/FDK.

    Parameters
    ----------

    slice_index : int, str, default='centre'
        An integer defining the vertical slice to run the algorithm on. The special case slice 'centre' is the default.

    backend : {'tigre', 'astra'}
        The backend to use for the reconstruction

    tolerance : float, default=0.005
        The tolerance of the fit in pixels, the default is 1/200 of a pixel. This is a stopping criteria, not a statement of accuracy of the algorithm.

    search_range : int
        The range in pixels to search either side of the panel centre. If `None` a quarter of the width of the panel is used.

    initial_binning : int
        The size of the bins for the initial search. If `None` will bin the image to a step corresponding to <128 pixels. The fine search will be on unbinned data.


    Example
    -------
    .. code-block :: python 
        from cil.processors import CentreOfRotationCorrector

        processor = CentreOfRotationCorrector.image_sharpness('centre', 'tigre')
        processor.set_input(data)
        data_centred = processor.get_output()


    Example
    -------
    .. code-block :: python
        from cil.processors import CentreOfRotationCorrector

        processor = CentreOfRotationCorrector.image_sharpness(slice_index=120, 'astra')
        processor.set_input(data)
        processor.get_output(out=data)


    Note
    ----
    For best results data should be 360deg which leads to blurring with incorrect geometry.
    This method is unreliable on half-scan data with 'tuning-fork' style artifacts.

    """
    _supported_backends = ['astra', 'tigre']

    def __init__(self, slice_index='centre', backend='tigre', tolerance=0.005, search_range=None, initial_binning=None):

        FBP = self._configure_FBP(backend)


        kwargs = {
                    'slice_index': slice_index,
                    'FBP': FBP,
                    'backend' : backend,
                    'tolerance': tolerance,
                    'search_range': search_range,
                    'initial_binning': initial_binning
                 }

        super(CofR_image_sharpness, self).__init__(**kwargs)

    def check_input(self, data):
        if not isinstance(data, AcquisitionData):
            raise Exception('Processor supports only AcquisitionData')

        if data.geometry == None:
            raise Exception('Geometry is not defined.')

        if data.geometry.geom_type == 'cone' and self.slice_index != 'centre':
            raise ValueError("Only the centre slice is supported with this algorithm")

        if data.geometry.geom_type & AcquisitionType.CONE_FLEX:
            raise ValueError("Cone-Flex geometry is not supported by this processor")
        
        if data.geometry.system_description not in ['simple','offset']:
            raise NotImplementedError("Not implemented for rotated system geometries")

        if data.geometry.channels > 1:
            raise ValueError("Only single channel data is supported with this algorithm")

        if self.slice_index != 'centre':
            try:
                int(self.slice_index)
            except:
                raise ValueError("slice_index expected to be a positive integer or the string 'centre'. Got {0}".format(self.slice_index))

            if self.slice_index < 0 or self.slice_index >= data.get_dimension_size('vertical'):
                raise ValueError('slice_index is out of range. Must be in range 0-{0}. Got {1}'.format(data.get_dimension_size('vertical'), self.slice_index))

        #check order for single slice data
        test_geom = data.geometry.get_slice(vertical='centre') if AcquisitionType.DIM3 & data.geometry.dimension else data.geometry

        if not AcquisitionDimension.check_order_for_engine(self.backend, test_geom):
            raise ValueError("Input data must be reordered for use with selected backend. Use input.reorder{'{0}')".format(self.backend))

        return True


    def _configure_FBP(self, backend='tigre'):
        """
        Configures the processor for the right engine. Checks the data order.
        """
        if backend not in self._supported_backends:
            raise ValueError("Backend unsupported. Supported backends: {}".format(self._supported_backends))

        #set FBPOperator class from backend
        module = importlib.import_module(f'cil.plugins.{backend}')

        return module.FBP


    def gss(self, data, ig, search_range, tolerance, binning):
        '''Golden section search'''
        # intervals c:cr:c where r = φ − 1=0.619... and c = 1 − r = 0.381..., φ
        log.debug("GSS between %f and %f", *search_range)
        phi = (1 + math.sqrt(5))*0.5
        r = phi - 1
        #1/(r+2)
        r2inv = 1/ (r+2)
        #c = 1 - r

        all_data = {}
        #set up
        sample_points = [np.nan]*4
        evaluation = [np.nan]*4

        sample_points[0] = search_range[0]
        sample_points[3] = search_range[1]

        interval = sample_points[-1] - sample_points[0]
        step_c = interval *r2inv
        sample_points[1] = search_range[0] + step_c
        sample_points[2] = search_range[1] - step_c

        for i in range(4):
            evaluation[i] = self.calculate(data, ig, sample_points[i])
            all_data[sample_points[i]] = evaluation[i]

        count = 0
        while(count < 30):
            ind = np.argmin(evaluation)
            if ind == 1:
                del sample_points[-1]
                del evaluation[-1]

                interval = sample_points[-1] - sample_points[0]
                step_c = interval *r2inv
                new_point = sample_points[0] + step_c

            elif ind == 2:
                del sample_points[0]
                del evaluation[0]

                interval = sample_points[-1] - sample_points[0]
                step_c = interval *r2inv
                new_point = sample_points[-1]- step_c

            else:
                raise ValueError("The centre of rotation could not be located to the requested tolerance. Try increasing the search tolerance.")

            if interval < tolerance:
                break

            sample_points.insert(ind, new_point)
            obj = self.calculate(data, ig, new_point)
            evaluation.insert(ind, obj)
            all_data[new_point] = obj

            count +=1

        log.info("evaluated %d points",len(all_data))
        if log.isEnabledFor(logging.DEBUG):
            keys, values = zip(*all_data.items())
            self.plot(keys, values, ig.voxel_size_x/binning)

        z = np.polyfit(sample_points, evaluation, 2)
        min_point = -z[1] / (2*z[0])

        if np.sign(z[0]) == 1 and min_point < sample_points[2] and min_point > sample_points[0]:
            return min_point
        else:
            ind = np.argmin(evaluation)
            return sample_points[ind]

    def calculate(self, data, ig, offset):
        ag_shift = data.geometry.copy()
        ag_shift.config.system.rotation_axis.position = [offset, 0]

        reco = self.FBP(ig, ag_shift)(data)
        return (reco*reco).sum()

    def plot(self, offsets,values, vox_size):
        import matplotlib.pyplot as plt

        x=[x / vox_size for x in offsets]
        y=values

        plt.figure()
        plt.scatter(x,y)
        plt.show()

    def get_min(self, offsets, values, ind):
        #calculate quadratic from 3 points around ind  (-1,0,1)
        a = (values[ind+1] + values[ind-1] - 2*values[ind]) * 0.5
        b = a + values[ind] - values[ind-1]
        ind_centre = -b / (2*a)+ind

        ind0 = int(ind_centre)
        w1 = ind_centre - ind0
        return (1.0 - w1) * offsets[ind0] + w1 * offsets[ind0+1]

    def process(self, out=None):
        #get slice
        data = data_full = self.get_input()

        if AcquisitionType.DIM3 & data_full.geometry.dimension:
            data = data.get_slice(vertical=self.slice_index)

        data.geometry.config.system.align_reference_frame('cil')
        width = data.geometry.config.panel.num_pixels[0]

        #initial grid search
        if self.search_range is None:
            self.search_range = width //4

        if self.initial_binning is None:
            self.initial_binning = min(int(np.ceil(width / 128)),16)

        log.debug("Initial search:")
        log.debug("search range is %d", self.search_range)
        log.debug("initial binning is %d", self.initial_binning)

        #filter full projections
        data_filtered = data.copy()
        data_filtered.fill(scipy.ndimage.sobel(data.as_array(), axis=1, mode='reflect', cval=0.0))

        if self.initial_binning > 1:

            #gaussian filter data
            data_temp = data_filtered.copy()
            data_temp.fill(scipy.ndimage.gaussian_filter(data_filtered.as_array(), [0,self.initial_binning//2]))

            #bin data whilst preserving centres
            num_pix_new = np.ceil(width/self.initial_binning)

            new_half_panel = (num_pix_new - 1)/2
            half_panel = (width - 1)/2

            sampling_points = np.mgrid[-self.initial_binning*new_half_panel:self.initial_binning*new_half_panel+1:self.initial_binning]
            initial_coordinates = np.mgrid[-half_panel:half_panel+1:1]

            new_geom = data.geometry.copy()
            new_geom.config.panel.num_pixels[0] = num_pix_new
            new_geom.config.panel.pixel_size[0] *= self.initial_binning
            data_binned = new_geom.allocate()

            for i in range(data.shape[0]):
                data_binned.fill(np.interp(sampling_points, initial_coordinates, data.array[i,:]),angle=i)

            #filter
            data_binned_filtered = data_binned.copy()
            data_binned_filtered.fill(scipy.ndimage.sobel(data_binned.as_array(), axis=1, mode='reflect', cval=0.0))
            data_processed = data_binned_filtered
        else:
            data_processed = data_filtered

        ig = data_processed.geometry.get_ImageGeometry()

        #binned grid search
        vox_rad = np.ceil(self.search_range /self.initial_binning)
        steps = int(4*vox_rad + 1)
        offsets = np.linspace(-vox_rad, vox_rad, steps) * ig.voxel_size_x
        obj_vals = []

        for offset in offsets:
            obj_vals.append(self.calculate(data_processed, ig, offset))

        if log.isEnabledFor(logging.DEBUG):
            self.plot(offsets,obj_vals,ig.voxel_size_x / self.initial_binning)

        ind = np.argmin(obj_vals)
        if ind == 0 or ind == len(obj_vals)-1:
            raise ValueError ("Unable to minimise function within set search_range")
        else:
            centre = self.get_min(offsets, obj_vals, ind)

        if self.initial_binning > 8:
            #binned search continued
            log.debug("binned search starting at %f", centre)
            a = centre - ig.voxel_size_x *2
            b = centre + ig.voxel_size_x *2
            centre = self.gss(data_processed,ig, (a, b), self.tolerance *ig.voxel_size_x, self.initial_binning )

        #fine search
        log.debug("fine search starting at %f", centre)
        data_processed = data_filtered
        ig = data_processed.geometry.get_ImageGeometry()
        a = centre - ig.voxel_size_x *2
        b = centre + ig.voxel_size_x *2
        centre = self.gss(data_processed,ig, (a, b), self.tolerance *ig.voxel_size_x, 1 )

        new_geometry = data_full.geometry.copy()
        new_geometry.config.system.rotation_axis.position[0] = centre

        log.info("Centre of rotation correction found using image_sharpness")
        log.info("backend FBP/FDK {}".format(self.backend))
        log.info("Calculated from slice: %s", str(self.slice_index))
        log.info("Centre of rotation shift = %f pixels", centre/ig.voxel_size_x)
        log.info("Centre of rotation shift = %f units at the object", centre)
        log.info("Return new dataset with centred geometry")

        if out is None:
            return AcquisitionData(array=data_full, deep_copy=True, geometry=new_geometry)
        else:
            out.geometry = new_geometry
            return out


#  Copyright 2024 United Kingdom Research and Innovation
#  Copyright 2024 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt
#
# This work has been supported by the Royal Academy of Engineering and the
# Office of the Chief Science Adviser for National Security under the UK
# Intelligence Community Postdoctoral Research Fellowship programme.
#
# Francis M Watson, University of Manchester 2024
#


import numpy as np
from cil.optimisation.functions import Function
from cil.framework import DataContainer
from typing import Optional
import warnings
import logging

log = logging.getLogger(__name__)

class FunctionOfAbs(Function):
    r'''A function which acts on the absolute value of the complex-valued input, 

    .. math::  G(z) = H(abs(z))

    This function is initialised with another CIL function, :math:`H` in the above formula. When this function is called, first the absolute value of the input is taken, and then the input is passed to the provided function.
    
    Included in this class is the proximal map for FunctionOfAbs.  From this, the proximal conjugate is also available from the parent CIL Function class, which is valid for this function. 
    In the case that :math:`H` is lower semi-continuous, convex, non-decreasing and finite at the origin, (and thus `assume_lower_semi` is set to `True` in the `init`) the convex conjugate is also defined.  
    The gradient is not defined for this function.
    
    For reference: see https://doi.org/10.48550/arXiv.2410.22161

    Parameters
    ----------
    function : Function
        Function acting on a real input, :math:`H` in the above formula.
    assume_lower_semi : bool, default False
        If True, assume that the function is lower semi-continuous, convex, non-decreasing and finite at the origin.
        This allows the convex conjugate to be calculated as the monotone conjugate, which is less than or equal to the convex conjugate.
        If False, the convex conjugate returned as 0. This is to ensure compatibility with Algorithms such as PDHG.

    '''

    def __init__(self, function: Function, assume_lower_semi: bool=False):
        self._function = function
        self._lower_semi = assume_lower_semi
                
        super().__init__(L=function.L)

    def __call__(self, x: DataContainer) -> float:
        call_abs = _take_abs_input(self._function.__call__)
        return call_abs(self._function, x)

    def proximal(self, x: DataContainer, tau: float, out: Optional[DataContainer]=None) -> DataContainer:
        r'''Returns the proximal map of function :math:`\tau G`  evaluated at x

        .. math:: \text{prox}_{\tau G}(x) = \underset{z}{\text{argmin}} \frac{1}{2}\|z - x\|^{2} + \tau G(z)

        This is accomplished by calculating a bounded proximal map and making a change of phase,
        :math:`prox_G(z) = prox^+_H(r) \circ \Phi` where  :math:`z = r \circ \Phi`, :math:`r = abs(z)`, :math:`\Phi = \exp(i \angle(z))`,
        and :math:`\circ` is element-wise product.  Also define :math:`prox^+` to be the proximal map of :math:`H`  in which the minimisation carried out over the positive orthant.


        Parameters
        ----------
        x : DataContainer
            The input to the function
        tau: scalar
            The scalar multiplying the function in the proximal map
        out: return DataContainer, if None a new DataContainer is returned, default None.
            DataContainer to store the result of the proximal map
        

        Returns
        -------
        DataContainer, the proximal map of the function at x with scalar :math:`\tau`.

        '''
        prox_abs = _abs_and_project(self._function.proximal)
        return prox_abs(self._function, x, tau=tau, out=out)

    def convex_conjugate(self, x: DataContainer) -> float:
        r'''
        Evaluation of the function G* at x, where G* is the convex conjugate of function G,

        .. math:: G^{*}(x^{*}) = \underset{x}{\sup} \langle x^{*}, x \rangle - G(x)

        If :math:`H` is lower semi-continuous, convex, non-decreasing 
        finite at the origin, then :math:`G^*(z*) = H^+(|z*|)`, where the monotone conjugate :math:`g^+` is

        .. math:: H^+(z^*) =sup {(z, z^*) - H(z) : z >= O}

        The monotone conjugate will therefore be less than or equal to the convex conjugate, 
        since it is taken over a smaller set.  It is not available directly, but may coincide with
        the convex conjugate, which is therefore the best estimate we have.  This is only valid for
        real x. In other cases, a general convex conjugate is not available or defined.      


        For reference see:  Convex Analysis, R. Tyrrell Rocakfellar, pp110-111.
        
        
        Parameters
        ----------
        x : DataContainer
            The input to the function

        Returns
        -------
        float:
        The value of the convex conjugate of the function at x.
        '''

        if self._lower_semi:
            conv_abs = _take_abs_input(self._function.convex_conjugate)
            return conv_abs(self._function, x)
        else:
            warnings.warn('Convex conjugate is not defined for this function, returning 0 for compatibility with optimisation algorithms')
            return 0.0

    def gradient(self, x):
        '''Gradient of the function at x is not defined for this function.
        '''
        raise NotImplementedError('Gradient not available for this function')




def _take_abs_input(func):
    '''Decorator for function to act on abs of input of a method'''

    def _take_abs_decorator(self, x: DataContainer, *args, **kwargs):
        real_dtype, _ = _get_real_complex_dtype(x)
        
        rgeo = x.geometry.copy()
        rgeo.dtype = real_dtype
        r = rgeo.allocate(0)
        r.fill(np.abs(x.as_array()).astype(real_dtype))
        fval = func(r, *args, **kwargs)
        return fval
    return _take_abs_decorator


def _abs_and_project(func):
    '''Decorator for function to act on abs of input, 
    with return being projected to the angle of the input.
    Requires function return to have the same shape as input,
    such as prox.'''
    
        
    def _abs_project_decorator(self, x: DataContainer, *args, **kwargs):
        
        real_dtype, complex_dtype = _get_real_complex_dtype(x)
            
                    
        rgeo = x.geometry.copy()
        rgeo.dtype = real_dtype
        r = rgeo.allocate(None)
        r.fill(np.abs(x.as_array()).astype(real_dtype))
        Phi = np.exp((1j*np.angle(x.array)))
        out = kwargs.pop('out', None)
        
        fvals = func(r, *args, **kwargs)
        fvals_np = fvals.as_array()
        
        # Douglas-Rachford splitting to find solution in positive orthant
        if np.any(fvals_np < 0):
            log.info('AbsFunctions: projection to +ve orthant triggered')
            cts = 0
            y = r.copy()
            fvals_np = fvals.as_array()
            while np.any(fvals_np < 0):
                tmp = fvals_np  - 0.5*y.as_array() + 0.5*r.as_array()
                tmp[tmp < 0] = 0.
                y += DataContainter(tmp, y.geometry) - fvals
                fvals = func(y, *args, **kwargs)
                cts += 1
                if cts > 10:
                    fvals_np = fvals.as_array()
                    fvals_np[fvals_np < 0] = 0.
                    break

        if out is None: 
            out_geom = x.geometry.copy()
            out = out_geom.allocate(None)
        if np.isreal(x.as_array()).all():
            out.fill( np.real(fvals_np.astype(complex_dtype)*Phi))
        else:
            out.fill( fvals_np.astype(complex_dtype)*Phi)
        return out
    return _abs_project_decorator


def _get_real_complex_dtype(x: DataContainer):
    '''An internal function to find the type of x and set the corresponding real and complex data types '''
    
    x_dtype = x.as_array().dtype
    if np.issubdtype(x_dtype, np.complexfloating):
        complex_dtype = x_dtype
        complex_example = np.array([1 + 1j], dtype=x_dtype)
        real_dtype = np.real(complex_example).dtype
    else: 
        real_dtype = x_dtype
        complex_example = 1j*np.array([1], dtype=x_dtype)
        complex_dtype = complex_example.dtype
    return real_dtype, complex_dtype            
            


#  Copyright 2018 United Kingdom Research and Innovation
#  Copyright 2018 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt


from cil.framework import DataProcessor, ImageData
from cil.plugins.astra.utilities import convert_geometry_to_astra_vec_2D
import astra
import numpy as np


class AstraBackProjector2D(DataProcessor):
    """
    AstraBackProjector2D configures an ASTRA 2D back projector for CPU and GPU.

    Parameters
    ----------

    volume_geometry : ImageGeometry
        A description of the area/volume to reconstruct

    sinogram_geometry : AcquisitionGeometry
        A description of the acquisition data

    proj_id : the ASTRA projector ID
        For advances ASTRA users only. The astra_mex_projector ID of the projector, use `astra.astra_create_projector()`

    device : string, default='gpu'
        The device to run on 'gpu' or 'cpu'

    """

    def __init__(self,
                 volume_geometry=None,
                 sinogram_geometry=None,
                 proj_id=None,
                 device='cpu'):
        kwargs = {
                  'volume_geometry'  : volume_geometry,
                  'sinogram_geometry'  : sinogram_geometry,
                  'proj_id'  : proj_id,
                  'device'  : device
                  }

        #DataProcessor.__init__(self, **kwargs)
        super(AstraBackProjector2D, self).__init__(**kwargs)

        self.set_ImageGeometry(volume_geometry)
        self.set_AcquisitionGeometry(sinogram_geometry)

        vol_geom, proj_geom = convert_geometry_to_astra_vec_2D(self.volume_geometry, self.sinogram_geometry)

        # ASTRA projector, to be stored
        if device == 'cpu':
            # Note that 'line' only one option
            if self.sinogram_geometry.geom_type == 'parallel':
                self.set_projector(astra.create_projector('line', proj_geom, vol_geom) )
            elif self.sinogram_geometry.geom_type == 'cone':
                self.set_projector(astra.create_projector('line_fanflat', proj_geom, vol_geom) )
            else:
                NotImplemented
        elif device == 'gpu':
            self.set_projector(astra.create_projector('cuda', proj_geom, vol_geom) )
        else:
            NotImplemented

    def check_input(self, dataset):
        if dataset.number_of_dimensions == 1 or dataset.number_of_dimensions == 2:
               return True
        else:
            raise ValueError("Expected input dimensions is 1 or 2, got {0}"\
                             .format(dataset.number_of_dimensions))
        
    def _set_up(self):
        """
        Configure processor attributes that require the data to setup
        Must set _shape_out
        """
        self._shape_out = self.volume_geometry.shape

    def set_projector(self, proj_id):
        self.proj_id = proj_id

    def set_ImageGeometry(self, volume_geometry):
        self.volume_geometry = volume_geometry

    def set_AcquisitionGeometry(self, sinogram_geometry):
        self.sinogram_geometry = sinogram_geometry

    def process(self, out=None):

        DATA = self.get_input()

        #ASTRA expects a 3D array with shape 1, CIL removes dimensions of len 1
        new_shape_ag = [self.sinogram_geometry.num_projections,self.sinogram_geometry.pixel_num_h]
        data_temp = DATA.as_array().reshape(new_shape_ag)

        rec_id, arr_out = astra.create_backprojection(data_temp, self.proj_id)
        astra.data2d.delete(rec_id)

        arr_out = np.squeeze(arr_out)
        if out is None:
            out = ImageData(arr_out, deep_copy=False, geometry=self.volume_geometry.copy())
        else:
            out.fill(arr_out)
        return out



#  Copyright 2025 United Kingdom Research and Innovation
#  Copyright 2025 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt


import unittest
import numpy as np
from cil.optimisation.algorithms.APGD import ConstantMomentum, NesterovMomentum, ScalarMomentumCoefficient

class MockAlgorithm:
    pass  # Placeholder for any required algorithm attributes

class TestMomentumCoefficients(unittest.TestCase):
    
    def test_constant_momentum(self):
        momentum_value = 0.9
        momentum = ConstantMomentum(momentum_value)
        algorithm = MockAlgorithm()
        
        self.assertEqual(momentum(algorithm), momentum_value, msg="ConstantMomentum should return the set momentum value")
    
    def test_nesterov_momentum(self):
        momentum = NesterovMomentum()
        algorithm = MockAlgorithm()
        
        initial_value = momentum(algorithm)
        second_value = momentum(algorithm)
        
        # Check initial momentum value
        expected_initial_t = 1
        expected_next_t = 0.5 * (1 + np.sqrt(1 + 4 * (expected_initial_t ** 2)))
        expected_initial_momentum = (expected_initial_t - 1) / expected_next_t
        
        self.assertAlmostEqual(initial_value, expected_initial_momentum, places=6, msg="Incorrect initial momentum value")
        
        # Check second iteration momentum value
        expected_next_t_old = expected_next_t
        expected_next_t = 0.5 * (1 + np.sqrt(1 + 4 * (expected_next_t_old ** 2)))
        expected_next_momentum = (expected_next_t_old - 1) / expected_next_t
        
        self.assertAlmostEqual(second_value, expected_next_momentum, places=6, msg="Incorrect second iteration momentum value")
        
    def test_momentum_coefficient_abc(self):
        class InvalidMomentum(ScalarMomentumCoefficient):
            pass
        
        with self.assertRaises(TypeError):
            InvalidMomentum()    


#  Copyright 2023 United Kingdom Research and Innovation
#  Copyright 2023 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

from cil.framework import AcquisitionData, ImageData, DataContainer
import os
from cil.io import utilities
import configparser

import logging

log = logging.getLogger(__name__)


def compress_and_save(data, compress, scale, offset, dtype, fname):
    '''Compress and save numpy array to file

    Parameters
    ----------
    data : numpy array
    compress : bool
    scale : float
    offset : float
    dtype : numpy dtype
    fname : string

    Note:
    -----
    Data is always written in ‘C’ order, independent of the order of d.
    '''
    if compress:
        d = utilities.compress_data(data, scale, offset, dtype)
    else:
        d = data

    log.info(
        "Data is always written in ‘C’ order, independent of the order of d.")
    d.tofile(fname)

    # return shape, fortran order, dtype
    return d.shape, False, d.dtype.str


class RAWFileWriter(object):
    '''
        Writer to write DataContainer (or subclass AcquisitionData, ImageData) to disk as a binary blob

        Parameters
        ----------
        data : DataContainer, AcquisitionData or ImageData
            This represents the data to save to TIFF file(s)
        file_name : string
            This defines the file name prefix, i.e. the file name without the extension.
        compression : str, default None. Accepted values None, 'uint8', 'uint16'
            The lossy compression to apply. The default None will not compress data.
            'uint8' or 'unit16' will compress to unsigned int 8 and 16 bit respectively.


        This writer will also write a text file with the minimal information necessary to
        read the data back in. This text file will need to reside in the same directory as the raw file.

        The text file will look something like this::

            [MINIMAL INFO]
            file_name = filename.raw
            data_type = <u2
            shape = (6, 5, 4)
            is_fortran = False

            [COMPRESSION]
            scale = 550.7142857142857
            offset = -0.0

        The ``data_type`` describes the data layout when packing and unpacking data. This can be
        read as numpy dtype with ``np.dtype('<u2')``.


        Example
        -------

        Example of using the writer with compression to ``uint8``:

        >>> from cil.io import RAWFileWriter
        >>> writer = RAWFileWriter(data=data, file_name=fname, compression='uint8')
        >>> writer.write()

        Example
        -------

        Example of reading the data from the ini file:

        >>> config = configparser.ConfigParser()
        >>> inifname = "file_name.ini"
        >>> config.read(inifname)
        >>> read_dtype = config['MINIMAL INFO']['data_type']
        >>> dtype = np.dtype(read_dtype)
        >>> fname = config['MINIMAL INFO']['file_name']
        >>> read_array = np.fromfile(fname, dtype=read_dtype)
        >>> read_shape = eval(config['MINIMAL INFO']['shape'])
        >>> scale = float(config['COMPRESSION']['scale'])
        >>> offset = float(config['COMPRESSION']['offset'])

        Note
        ----

          If compression ``uint8`` or ``unit16`` are used, the scale and offset used to compress the data are saved
          in the ``ini`` file in the same directory as the raw file, in the "COMPRESSION" section .

          The original data can be obtained by: ``original_data = (compressed_data - offset) / scale``

        Note
        ----

          Data is always written in ‘C’ order independent of the order of the original data,
          https://numpy.org/doc/stable/reference/generated/numpy.ndarray.tofile.html#numpy.ndarray.tofile,


    '''

    def __init__(self, data, file_name, compression=None):

        if not isinstance(data, (DataContainer, ImageData, AcquisitionData)):
            raise Exception('Writer supports only following data types:\n' +
                            'DataContainer - ImageData\n - AcquisitionData')

        self.data_container = data
        file_name = os.path.abspath(file_name)
        self.file_name = os.path.splitext(os.path.basename(file_name))[0]

        self.dir_name = os.path.dirname(file_name)
        log.info("dir_name %s", self.dir_name)
        log.info("file_name %s", self.file_name)

        # Deal with compression
        self.compress = utilities.get_compress(compression)
        self.dtype = utilities.get_compressed_dtype(data, compression)
        self.scale, self.offset = utilities.get_compression_scale_offset(
            data, compression)
        self.compression = compression

    def write(self):
        '''Write data to disk'''
        if not os.path.isdir(self.dir_name):
            os.mkdir(self.dir_name)

        fname = os.path.join(self.dir_name, self.file_name + '.raw')

        # write to disk
        header = \
            compress_and_save(self.data_container.as_array(), self.compress, self.scale, self.offset, self.dtype, fname)

        shape = header[0]
        fortran_order = header[1]
        read_dtype = header[2]

        # save information about the file we just saved
        config = configparser.ConfigParser()
        config['MINIMAL INFO'] = {
            'file_name': os.path.basename(fname),
            'data_type': read_dtype,
            'shape': shape,
            # Data is always written in ‘C’ order, independent of the order of d.
            'is_fortran': fortran_order
        }

        if self.compress:
            config['COMPRESSION'] = {
                'scale': self.scale,
                'offset': self.offset,
            }
        log.info("Saving to %s", self.file_name)
        log.info(str(config))
        # write the configuration to an ini file
        with open(os.path.join(self.dir_name, self.file_name + '.ini'),
                  'w') as configfile:
            config.write(configfile)



#   Copyright 2024 United Kingdom Research and Innovation
#   Copyright 2024 The University of Manchester
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#  Authors:
#  - CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt
#  - Daniel Deidda (National Physical Laboratory, UK)
#  - Claire Delplancke (Electricite de France, Research and Development)
#  - Ashley Gillman (Australian e-Health Res. Ctr., CSIRO, Brisbane, Queensland, Australia)
#  - Zeljko Kereta (Department of Computer Science, University College London, UK)
#  - Evgueni Ovtchinnikov (STFC - UKRI)
#  - Georg Schramm (Department of Imaging and Pathology, Division of Nuclear Medicine, KU Leuven, Leuven, Belgium)


from .ApproximateGradientSumFunction import ApproximateGradientSumFunction
import numpy as np
import numbers


class SVRGFunction(ApproximateGradientSumFunction):

    r"""
    The Stochastic Variance Reduced Gradient (SVRG) function calculates the approximate gradient of :math:`\sum_{i=1}^{n-1}f_i`.  For this approximation, every `snapshot_update_interval` number of iterations, a full gradient calculation is made at this "snapshot" point. Intermediate gradient calculations update this snapshot by taking a index :math:`i_k` and calculating the gradient of :math:`f_{i_k}`'s at the current iterate and the snapshot, updating the approximate gradient to be:

    .. math ::
        n*\nabla f_{i_k}(x_k) - n*\nabla f_{i_k}(\tilde{x}) + \nabla \sum_{i=0}^{n-1}f_i(\tilde{x}),

    where :math:`\tilde{x}` is the latest "snapshot" point and :math:`x_k` is the value at the current iteration. 

    Note
    -----
    Compared with the literature, we multiply by :math:`n`, the number of functions, so that we return an approximate gradient of the whole sum function and not an average gradient. 

    Note
    ----
    In the case where `store_gradients=False` the memory requirements are 4 times the image size (1 stored full gradient at the "snapshot", one stored "snapshot" point and two lots of intermediary calculations). Alternatively, if  `store_gradients=True`  the memory requirement is `n+4` (`n` gradients at the snapshot for each function in the sum, one stored full gradient at the "snapshot", one stored "snapshot" point and two lots of intermediary calculations).
    
    Reference
    ---------
    Johnson, R. and Zhang, T., 2013. Accelerating stochastic gradient descent using predictive variance reduction. Advances in neural information processing systems, 26.https://proceedings.neurips.cc/paper_files/paper/2013/file/ac1dd209cbcc5e5d1c6e28598e8cbbe8-Paper.pdf


    Parameters
    ----------
    functions : `list`  of functions
        A list of functions: :code:`[f_{0}, f_{1}, ..., f_{n-1}]`. Each function is assumed to be smooth with an implemented :func:`~Function.gradient` method. All functions must have the same domain. The number of functions must be strictly greater than 1. 
    sampler: An instance of a CIL Sampler class ( :meth:`~optimisation.utilities.sampler`) or of another class which has a `next` function implemented to output integers in {0, 1, ..., n-1}.
        This sampler is called each time gradient is called and  sets the internal `function_num` passed to the `approximate_gradient` function.  Default is `Sampler.random_with_replacement(len(functions))`. 
    snapshot_update_interval : positive int or None, optional
        The interval for updating the full gradient (taking a snapshot). The default is 2*len(functions) so a "snapshot" is taken every 2*len(functions) iterations. If the user passes `0` then no full gradient snapshots will be taken. 
    store_gradients : bool, default: `False`
        Flag indicating whether to store and update a list of gradients for each function :math:`f_i` or just to store the snapshot point :math:`\tilde{x}` and its gradient :math:`\nabla \sum_{i=0}^{n-1}f_i(\tilde{x})`.


    """

    def __init__(self, functions, sampler=None, snapshot_update_interval=None, store_gradients=False):
        super(SVRGFunction, self).__init__(functions, sampler)

        #  snapshot_update_interval for SVRG
        self.snapshot_update_interval = snapshot_update_interval
    
        if snapshot_update_interval is None:
            self.snapshot_update_interval = 2*self.num_functions
        self.store_gradients = store_gradients

        self._svrg_iter_number = 0

        self._full_gradient_at_snapshot = None
        self._list_stored_gradients = None

        self.stoch_grad_at_iterate = None
        self._stochastic_grad_difference = None
        
        self.snapshot = None

    def gradient(self, x, out=None):
        r""" Selects a random function using the `sampler` and then calls the approximate gradient at :code:`x` or calculates a full gradient depending on the update frequency

        Parameters
        ----------
        x : DataContainer (e.g. ImageData object)
        out: return DataContainer, if `None` a new DataContainer is returned, default `None`.

        Returns
        --------
        DataContainer (e.g. ImageData object)
            the value of the approximate gradient of the sum function at :code:`x` 
        """

        #  For SVRG, every `snapshot_update_interval` a full gradient step is calculated, else an approximate gradient is taken.
        if ( (self.snapshot_update_interval != 0) and (self._svrg_iter_number % (self.snapshot_update_interval)) == 0):

            return self._update_full_gradient_and_return(x, out=out)

        else:

            self.function_num = self.sampler.next()
            if not isinstance(self.function_num, numbers.Number):
                raise ValueError("Batch gradient is not yet implemented")
            if self.function_num >= self.num_functions or self.function_num < 0:
                raise IndexError(
                    f"The sampler has produced the index {self.function_num} which does not match the expected range of available functions to sample from. Please ensure your sampler only selects from [0,1,...,len(functions)-1] ")
            return self.approximate_gradient(x, self.function_num, out=out)

    def approximate_gradient(self, x, function_num, out=None):
        r""" Calculates the stochastic gradient at the point :math:`x` by using the gradient of the selected function, indexed by :math:`i_k`, the `function_number` in {0,...,len(functions)-1}, and the full gradient at the snapshot :math:`\tilde{x}`
        
        .. math ::
            n*\nabla f_{i_k}(x_k) - n*\nabla f_{i_k}(\tilde{x}) + \nabla \sum_{i=0}^{n-1}f_i(\tilde{x})

        Note
        -----
        Compared with the literature, we multiply by :math:`n`, the number of functions, so that we return an approximate gradient of the whole sum function and not an average gradient.

        Parameters
        ----------
        x : DataContainer (e.g. ImageData object)
        out: return DataContainer, if `None` a new DataContainer is returned, default `None`.
        function_num: `int` 
            Between 0 and n-1, where n is the number of functions in the list  
        Returns
        --------
        DataContainer (e.g. ImageData object)
            the value of the approximate gradient of the sum function at :code:`x` given a `function_number` in {0,...,len(functions)-1}
        """

        self._svrg_iter_number += 1

        self.stoch_grad_at_iterate = self.functions[function_num].gradient(x, out=self.stoch_grad_at_iterate)

        if self.store_gradients is True:
            self._stochastic_grad_difference = self.stoch_grad_at_iterate.sapyb(
                1., self._list_stored_gradients[function_num], -1., out=self._stochastic_grad_difference)
        else:
            self._stochastic_grad_difference = self.stoch_grad_at_iterate.sapyb(
                1., self.functions[function_num].gradient(self.snapshot), -1., out=self._stochastic_grad_difference)

        self._update_data_passes_indices([function_num])

        out = self._stochastic_grad_difference.sapyb(
            self.num_functions, self._full_gradient_at_snapshot, 1., out=out)

        return out

    def _update_full_gradient_and_return(self, x, out=None):
        """
        Takes a "snapshot" at the point :math:`x`, saving both the point :math:`\tilde{x}=x` and its gradient :math:`\sum_{i=0}^{n-1}f_i{\tilde{x}}`. The function returns :math:`\sum_{i=0}^{n-1}f_i{\tilde{x}}` as the gradient calculation. If :code:`store_gradients==True`, the gradient of all the :math:`f_i`s is computed and stored at the "snapshot"..

        Parameters
        ----------
        Takes a "snapshot" at the point :math:`x`. The function returns :math:`\sum_{i=0}^{n-1}f_i{\tilde{x}}` as the gradient calculation. If :code:`store_gradients==True`, the gradient of all the :math:`f_i`s is stored, otherwise only  the sum of the gradients and the snapshot point :math:`\tilde{x}=x` are stored.
        out: return DataContainer, if `None` a new DataContainer is returned, default `None`.

        Returns
        --------
        DataContainer (e.g. ImageData object)
            the value of the approximate gradient of the sum function at :code:`x` given a `function_number` in {0,...,len(functions)-1}
        """

        self._svrg_iter_number += 1

        if self.store_gradients is True:
            if self._list_stored_gradients is None:
                # Save the gradient of each individual f_i and the gradient of the full sum at the point x.
                self._list_stored_gradients = [
                    fi.gradient(x) for fi in self.functions]
                self._full_gradient_at_snapshot = sum(
                    self._list_stored_gradients, start=0*x)
            else:
                for i, fi in enumerate(self.functions):
                    fi.gradient(x, out=self._list_stored_gradients[i])

                self._full_gradient_at_snapshot.fill(
                    sum(self._list_stored_gradients, start=0*x))
                self._full_gradient_at_snapshot *= 0

                for i, el in enumerate(self._list_stored_gradients):
                    self._full_gradient_at_snapshot += el

        else:
            # Save the snapshot point and the gradient of the full sum at the point x.
            self._full_gradient_at_snapshot = self.full_gradient(
                x, out=self._full_gradient_at_snapshot)

        if self.snapshot is None:
            self.snapshot = x.copy()

        self.snapshot.fill(x)

        # In this iteration all functions in the sum were used to update the gradient
        self._update_data_passes_indices(list(range(self.num_functions)))

        # Return the gradient of the full sum at the snapshot.
        if out is None:
            out = self._full_gradient_at_snapshot
        else:
            out.fill(self._full_gradient_at_snapshot)

        return out


class LSVRGFunction(SVRGFunction):
    r"""
    A class representing a function for Loopless Stochastic Variance Reduced Gradient (SVRG) approximation. This is similar to SVRG, except the full gradient at a "snapshot"  is calculated at random intervals rather than at fixed numbers of iterations. 
    
    Parameters
    ----------
    functions : `list`  of functions
        A list of functions: :code:`[f_{0}, f_{1}, ..., f_{n-1}]`. Each function is assumed to be smooth with an implemented :func:`~Function.gradient` method. All functions must have the same domain. The number of functions `n` must be strictly greater than 1. 
    sampler: An instance of a CIL Sampler class ( :meth:`~optimisation.utilities.sampler`) or of another class which has a `next` function implemented to output integers in {0,...,n-1}.
        This sampler is called each time gradient is called and  sets the internal `function_num` passed to the `approximate_gradient` function.  Default is `Sampler.random_with_replacement(len(functions))`. 
    snapshot_update_probability: positive float, default: 1/n
        The probability of updating the full gradient (taking a snapshot) at each iteration. The default is :math:`1./n` so, in expectation, a snapshot will be taken every :math:`n` iterations. 
    store_gradients : bool, default: `False`
        Flag indicating whether to store and update a list of gradients for each function :math:`f_i` or just to store the snapshot point :math:`\tilde{x}` and it's gradient :math:`\nabla \sum_{i=0}^{n-1}f_i(\tilde{x})`.
    seed: int
        Seed for calculating the random update snapshot intervals (generated using numpy.random).

        
    Reference
    ---------
    Kovalev, D., Horváth, S. &; Richtárik, P.. (2020). Don’t Jump Through Hoops and Remove Those Loops:  SVRG and Katyusha are Better Without the Outer Loop. Proceedings of the 31st International Conference  on Algorithmic Learning Theory, in Proceedings of Machine Learning Research 117:451-467 Available from https://proceedings.mlr.press/v117/kovalev20a.html.


    Note
    ----
    In the case where `store_gradients=False` the memory requirements are 4 times the image size (1 stored full gradient at the "snapshot", one stored "snapshot" point and two lots of intermediary calculations). Alternatively, if  `store_gradients=True`  the memory requirement is `n+4` (`n` gradients at the snapshot for each function in the sum, one stored full gradient at the "snapshot", one stored "snapshot" point and two lots of intermediary calculations).

    """

    def __init__(self, functions, sampler=None, snapshot_update_probability=None, store_gradients=False, seed=None):

        super(LSVRGFunction, self).__init__(
            functions, sampler=sampler, store_gradients=store_gradients)

        # Update frequency based on probability.
        self.snapshot_update_probability = snapshot_update_probability
        #  Default snapshot_update_probability for Loopless SVRG
        if self.snapshot_update_probability is None:
            self.snapshot_update_probability = 1./self.num_functions

        #  The random generator used to decide if the gradient calculation is a full gradient or an approximate gradient
        self.generator = np.random.default_rng(seed=seed)

    def gradient(self, x, out=None):
        """ Selects a random function using the `sampler` and then calls the approximate gradient at :code:`x` or calculates a full gradient depending on the update probability.

        Parameters
        ----------
        x : DataContainer (e.g. ImageData objects)
        out: return DataContainer, if `None` a new DataContainer is returned, default `None`.

        Returns
        --------
        DataContainer (e.g. ImageData object)
            the value of the approximate gradient of the sum function at :code:`x`
        """

        if self._svrg_iter_number == 0 or self.generator.uniform() < self.snapshot_update_probability:

            return self._update_full_gradient_and_return(x, out=out)

        else:

            self.function_num = self.sampler.next()
            if not isinstance(self.function_num, numbers.Number):
                raise ValueError("Batch gradient is not yet implemented")
            if self.function_num >= self.num_functions or self.function_num < 0:
                raise IndexError(
                    f"The sampler has produced the index {self.function_num} which does not match the expected range of available functions to sample from. Please ensure your sampler only selects from [0,1,...,len(functions)-1] ")
            return self.approximate_gradient(x, self.function_num, out=out)



#  Copyright 2019 United Kingdom Research and Innovation
#  Copyright 2019 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

import numpy
from cil.optimisation.algorithms import Algorithm
import logging
from cil.optimisation.utilities import ConstantStepSize, ArmijoStepSizeRule, StepSizeRule
from warnings import warn
from numbers import Real

log = logging.getLogger(__name__)


class GD(Algorithm):
    """Gradient Descent algorithm

    Parameters
    ----------
    initial: DataContainer (e.g. ImageData)
        The initial point for the optimisation 
    f: CIL function (:meth:`~cil.optimisation.functions.Function`. ) with a defined gradient method 
        The function to be minimised. 
    step_size: positive real float or subclass of :meth:`~cil.optimisation.utilities.StepSizeRule`, default = None 
        If you pass a float this will be used as a constant step size. If left as None and do not pass a step size rule then the Armijo rule will be used to perform backtracking to choose a step size at each iteration (:meth:`~cil.optimisation.utilities.ArmijoStepSizeRule`). If a child class of :meth:`cil.optimisation.utilities.StepSizeRule`' is passed then its method `get_step_size` is called for each update. 
    preconditioner: class with a `apply` method or a function that takes an initialised CIL function as an argument and modifies a provided `gradient`.
            This could be a custom `preconditioner` or one provided in :meth:`~cil.optimisation.utilities.preconditioner`. If None is passed  then `self.gradient_update` will remain unmodified. 

    """


    def __init__(self, initial=None, f=None, step_size=None,   preconditioner=None, **kwargs):

        self.alpha = kwargs.pop('alpha', None) # to be deprecated
        self.beta = kwargs.pop('beta', None) # to be deprecated
        self.rtol = kwargs.pop('rtol', None) # to be deprecated (released in 25.0)
        self.atol = kwargs.pop('atol', None) # to be deprecated (released in 25.0)
        

        if kwargs.get('objective_function') is not None: # to be deprecated
            warn('The argument `objective_function` will be deprecated in the future. Please use `f` instead.', DeprecationWarning, stacklevel=2)
            if f is not None:
                raise ValueError('The argument `objective_function` is being deprecated, replaced by `f`. Please use just `f` not both')
            f = kwargs.pop('objective_function')
            

        if self.alpha is not None or self.beta is not None: # to be deprecated
            warn('To modify the parameters for the Armijo rule please use `step_size_rule=ArmijoStepSizeRule(alpha, beta, kmax)`. The arguments `alpha` and `beta` will be deprecated. ', DeprecationWarning, stacklevel=2)

        if self.rtol is not None or self.atol is not None: # to be deprecated (released in 25.0)
            if self.rtol != 0 or self.atol != 0: # to be deprecated (released in 25.0)
                warn('`rtol` and `atol` are deprecated. For early stopping, please use a callback (cil.optimisation.utilities.callbacks), for example `EarlyStoppingObjectiveValue`.', DeprecationWarning, stacklevel=2)
            else:
                log.warning('Breaking backwards compatibility, GD no longer automatically stops if the objective function is close to zero. For this functionality, please use a callback (cil.optimisation.utilities.callbacks).')
            
        super().__init__(**kwargs)
        
        if initial is not None and f is not None:
            self.set_up(initial=initial, f=f, step_size=step_size,  preconditioner=preconditioner)

    def set_up(self, initial, f, step_size, preconditioner):
        '''initialisation of the algorithm

        Parameters
        ----------
        initial: DataContainer (e.g. ImageData)
            The initial point for the optimisation 
        f: CIL function with a defined gradient 
            The function to be minimised. 
        step_size: positive real float or subclass of :meth:`~cil.optimisation.utilities.StepSizeRule`, default = None 
            If you pass a float this will be used as a constant step size. If left as None and do not pass a step_size_rule then the Armijio rule will be used to perform backtracking to choose a step size at each iteration. If a child class of :meth:`cil.optimisation.utilities.StepSizeRule`' is passed then it's method `get_step_size` is called for each update. 
        preconditioner: class with a `apply` method or a function that takes an initialised CIL function as an argument and modifies a provided `gradient`.
            This could be a custom `preconditioner` or one provided in :meth:`~cil.optimisation.utilities.preconditioner`. If None is passed  then `self.gradient_update` will remain unmodified. 

        '''

        log.info("%s setting up", self.__class__.__name__)

        self.x = initial.copy()
        self._objective_function = f

        if step_size is None:
            self.step_size_rule = ArmijoStepSizeRule(
                alpha=self.alpha, beta=self.beta)
        elif isinstance(step_size, Real):
            self.step_size_rule = ConstantStepSize(step_size)
        elif isinstance(step_size, StepSizeRule):
            self.step_size_rule = step_size
        else:
            raise TypeError(
                '`step_size` must be `None`, a Real float or a child class of :meth:`cil.optimisation.utilities.StepSizeRule`')
        self.gradient_update = initial.copy()

        self.configured = True

        self.preconditioner = preconditioner

        log.info("%s configured", self.__class__.__name__)

    def update(self):
        '''Performs a single iteration of the gradient descent algorithm'''
        self._objective_function.gradient(self.x, out=self.gradient_update)

        if self.preconditioner is not None:
            self.preconditioner.apply(
                self, self.gradient_update, out=self.gradient_update)

        self._step_size = self.step_size_rule.get_step_size(self)

        self.x.sapyb(1.0, self.gradient_update, -self._step_size, out=self.x)

    def update_objective(self):
        self.loss.append(self._objective_function(self.solution))

    def should_stop(self): # to be deprecated (released in 25.0)
        '''Stopping criterion for the gradient descent algorithm '''
        check_should_stop = False 
        if self.rtol  is not None or self.atol is not None:
            check_should_stop = numpy.isclose(self.get_last_objective(), 0., rtol=self.rtol,
                          atol=self.atol, equal_nan=False)
        return super().should_stop() or  check_should_stop
            

    @property
    def step_size(self):
        '''
        Returns the most recently used step size. Note, if the step-size is set by a non-constant step size rule, you must use the algorithm run or update method before this getter will return the most recently used step size. 
        '''
        
        if isinstance(self.step_size_rule, ConstantStepSize):
            return self.step_size_rule.step_size
        else:
            try: 
                return self._step_size
            except AttributeError:
                raise NotImplementedError("Note the step-size is set by a step-size rule and could change with each iteration. Call the algorithm run or update method first and then this function will give the most recently used step size.")


    def calculate_objective_function_at_point(self, x):
        """ Calculates the objective at a given point x

        .. math:: f(x)
        
        Parameters
        ----------
        x : DataContainer
        
        """
        return self._objective_function(x)
    
    @property
    def objective_function(self):
        warn('The attribute `objective_function` will be deprecated in the future. Please use `calculate_objective_function_at_point` instead.', DeprecationWarning, stacklevel=2)  
        return self._objective_function